/**
 * Name:  binary_search.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <pthread.h>
#include <assert.h>

#include "common.h"

#define THREADS_MAX   (128)
#define INDEX_MEM_MAX (1024*1024*1024)

//Structure for holding index data
typedef struct {
  u64    num_index_vals;
  u64    vects_per_index;
  u64    vect_length;
  u64    file_vects;
  float *data;
} index_t;

typedef struct {
  int   id;
  u64   num_threads;
  u64   total_input_vectors;
  char *fn_unsorted;
  char *fn_sorted;
  char *counted;
  u64   hit;
  u64   miss;
  index_t *index_block;
} thread_info_t;

int dump_statistics=0;  //Nonzero when print out as go along
int clear_cache=1;      //Clear system cache between runs

//=============================================================================
void dumpHelp(void){
  printf("binary_search <options> [-u unsorted_input_file]  sorted_file\n"
	 "\n"
	 "    -t/T threads        : min/max threads to use to check different inputs at same time\n"
	 "    -b/B index_buf_size : min/max bytes (if any) to allocate for an in-memory index\n"
	 "    -n max_vectors      : only check first max_vectors\n"
	 "    -v                  : print periodic summaries\n"
	 "    -C                  : disable the clearing of cache between iterations\n"
	 "    -d                  : print debug info\n"
	 "    -u unsorted_file    : supply inputs from unsorted file (instead of randomly generating)\n"
	 "    -q                  : print output in tabular form\n"
	 "    -Q                  : print header for tables\n"
	 "    -X                  : exit before doing anything\n"
	 "    sorted_file         : file that is sorted\n"
	 "\n");
  exit(0);
}


//=============================================================================
static int cmpVectors(const void *x, const void *y, u64 vect_length){
  float *xx, *yy;
  int i;

  xx = (float *)x;
  yy = (float *)y;

  for(i=0; i<vect_length;i++){
    if(xx[i] < yy[i]) return -1;
    if(xx[i] > yy[i]) return  1;
    //Equals, keep going
  }
  return 0; //All were equal
}



//=============================================================================
static void createIndex(char *filename, u64 tot_index_mem,
		 index_t **iptr){

  FILE *f;
  index_t *x;
  u64 tmp_len, tmp_num, tmp_type;
  u64 i;
  size_t rc;
  u64 bytes_per_vector;
  u64 num_index_vals;
 

  f = openFileWithMetaOrDie(filename, "r", &tmp_len, &tmp_num, &tmp_type);
  x = (index_t *)callocOrDie(1,sizeof(index_t));
  x->file_vects  = tmp_num;
  x->vect_length = tmp_len;

  //Bail out if didn't ask for any index cache
  if(!tot_index_mem){
    fclose(f);
    *iptr = x;
    return;
  }

  //Allocate the index buffer space
  bytes_per_vector = tmp_len*getVTypeSize(tmp_type);
  num_index_vals   = tot_index_mem  / bytes_per_vector;
  num_index_vals   = (num_index_vals>tmp_num) ? tmp_num : num_index_vals; //limit
  tot_index_mem    = num_index_vals * bytes_per_vector;


  x->num_index_vals  = num_index_vals;
  x->vects_per_index = tmp_num / num_index_vals;
  x->file_vects      = tmp_num;
  x->vect_length     = tmp_len;
  x->data = (float *) mallocOrDie(tot_index_mem);
  assert(num_index_vals>4); //Probably an error if don't have many slots


  //Read all the index vals
  for(i=0; i<num_index_vals; i++){

    rc = fread( &x->data[i * x->vect_length], 
		bytes_per_vector,
		1,  
		f);

    if(rc!=1)
      emsg("Bad read on indexing?");


    //Skip ahead to next block
    if(i!=num_index_vals-1){ //ignore last
      rc = fseeko(f,(x->vects_per_index-1)*bytes_per_vector,SEEK_CUR);
      if(rc)
	emsg("Problem seeking during index");
    }   
  }

  //printf("Stored %d index values\n", num_index_vals);

  fclose(f);
  *iptr = x;
  return;
}

//=============================================================================
// note: always want to return a range to search
static int getInitialStartStop(index_t *iptr, 
			float *input_vector,
			u64 *top, u64 *bot){

  u64 itop, ibot, imid;
  int cmp;
  int diff;

  //Bail out with defaults if not using indexing
  if(iptr->num_index_vals==0){
    *top  = 0;
    *bot  = iptr->file_vects-1;
    return 1;
  }

  //printf("Starting\n");
  //Binary search on memory structure to find best starting place
  itop = 0;
  ibot = iptr->num_index_vals-1;

  do{


    //if(ibot<itop){
    //  printf("Top/Bot/Mid: %llu/%llu/%llu cmp:%d\n", itop,ibot,imid, cmp);
      //}

    assert(ibot>=itop);

    diff = (ibot-itop);
    imid = itop + (diff/2);


    cmp = cmpVectors(input_vector, &iptr->data[ imid * iptr->vect_length ], iptr->vect_length);

    //printf("Top/Bot/Mid: %llu/%llu/%llu cmp:%d diff: %d", itop,ibot,imid, cmp,diff);
    //{ int j;
    //  for(j=0; j<iptr->vect_length; j++)
    // 	printf("(%.3lf/%.3lf) ", iptr->data[ imid * iptr->vect_length +j ], input_vector[j]);
    //  printf("\n");
    //}


    if(!cmp){
      //Found exact hit, don't update
      ibot = itop = imid;
    } else {

      if(diff>1){
	//Top/bottom are still different and can be updated
	if      (cmp<0) ibot = imid - 1; //input<midpoint : move bottom clamp up
	else if (cmp>0) itop = imid + 1; //input>mindpoint: move top clamp down
      } else if(diff==1){
	//Two spots, just checked lower
	if(cmp<0) diff=0; //Less than lower: no value works, so bail
	else      itop++; //Still possible for upper to have it
      } //diff=0: just pass back compare
    }

  } while(cmp && diff);


  //Need to return a range of one block of vectors
  if(cmp<0){
    //Value is before top
    if(itop){
      //Less than current block, but not first block
      *top = (itop-1) * iptr->vects_per_index;
      *bot = (itop    * iptr->vects_per_index) - 1;
    } else {
      //Less than the first entry..problem
      *top = 0;
      *bot = 0;
    }
  } else if (cmp>0){
    //Larger than first node
    *top =   itop    * iptr->vects_per_index;
    *bot = ((itop+1) * iptr->vects_per_index)-1;

    //Last index: need to clamp to last vector in file (because buffer likely to be short)
    if(itop==iptr->num_index_vals-1)
      *bot = iptr->file_vects-1;


  } else {
    //direct hit
    *bot = *top = ibot*iptr->vects_per_index;
  }



  return cmp;
}



//=============================================================================
void * th_cmp(void *arg){

  thread_info_t *ti;
  FILE *fs,*fu;
  char *count=NULL;
  u64 vlen, vnum, vtype;
  float *v1,*v2;
  u64 vecs_to_do, vecs_per_thread;
  s64 rc;
  u64 top,mid,bot;
  int cmp;
  u64 i,j;
  unsigned int seed;
  index_t *index;
  int diff;
  int exact_hit=0;

  struct timeval tv[2];
  int tv_pp=0;

  ti = (thread_info_t *)arg;

  //Open sorted and grab vector params from it
  fs = openFileWithMetaOrDie(ti->fn_sorted,"r", &vlen, &vnum, &vtype);
  if(!fs) emsg("failed to open sorted file");

  //Figure out how much work there is for us
  vecs_to_do = vecs_per_thread = ti->total_input_vectors/ti->num_threads;
  if(ti->id == ti->num_threads-1) //Last one gets extras
    vecs_to_do += (ti->total_input_vectors % ti->num_threads);


  //Open the unsorted file
  if(ti->fn_unsorted){
    //Reading from unsorted file
    fu = openFileWithMetaOrDie(ti->fn_unsorted, "r", NULL,  NULL,  NULL);
    if(!fu) emsg("failed to open unsorted file");

    rc = fseeko(fu, file_config_bytes + (ti->id*vecs_per_thread*vlen*sizeof(float)), SEEK_SET);
    if(rc<0) emsg("failed to seek");

  } else {
    //No input file, we generate input vector on fly
    fu = NULL;
    seed = ti->id;
  }


  dbg("Thread %d: vlen=%llu vnum=%llu  vecs/thd=%llu  vecs_to_do=%llu\n", 
      ti->id, vlen, vnum, vecs_per_thread, vecs_to_do);
  

  //Allocate space for everything
  v1 = (float *)mallocOrDie(vlen*sizeof(float));
  v2 = (float *)mallocOrDie(vlen*sizeof(float));



  gettimeofday(&tv[0],NULL);

  //Run through all unsorted vectors in our section
  for(i=0;i<vecs_to_do;i++){

    //Come up with next input vector (from file or generate)
    if(fu){
      rc = fread(v2, sizeof(float), vlen, fu);
      if(rc!=vlen) emsg("Bad read of input vector");
    } else {
      for(j=0; j<vlen; j++)
	v2[j] =  ((float)rand_r(&seed)) / ((float)RAND_MAX);
    }


    //Get start and stop from cache
    cmp = getInitialStartStop(ti->index_block, v2, &top, &bot);
    

    if(cmp){ //This allows us to skip out if initial start/stop found exact value
      //top = 0;
      //bot = vnum-1;
      assert(top>=0);
      assert(top<vnum);

      do{
	
	diff = (bot-top);
	mid = top + (diff/2);
	dbg("Testing %d-%d-%d",top, mid, bot);
	
	rc = fseeko(fs, file_config_bytes + (mid*vlen*sizeof(float)),SEEK_SET);
	if(rc<0) emsg("failed to seek");
	
	rc = fread(v1, sizeof(float), vlen, fs);
	if(rc!=vlen) emsg("failed to read");
	if(debug){  printf("V1(%d):",mid); dumpVector(v1, vlen, VT_F32, 0); printf("V2(%d):",i); dumpVector(v2, vlen, VT_F32, 0);}
	

	cmp = cmpVectors( v2, v1, vlen );
	
	if(!cmp){ //found exact hit
	  bot=top=mid;
	} else {

	  if(diff>1){
	    //Top/bottom are still different and can be updated
	    if      (cmp<0) bot = mid - 1; //input<midpoint : move bottom clamp up
	    else if (cmp>0) top = mid + 1; //input>mindpoint: move top clamp down
	  } else if(diff==1){
	    //Two spots, just checked lower
	    if(cmp<0) diff=0; //Less than lower: no value works, so bail
	    else      top++; //Still possible for upper to have it
	  } //diff=0: just pass back compare
	}
	
      } while(cmp && diff);
    } else {
      exact_hit++;
    }

    if(cmp){
      ti->miss++;
      //printf("Miss at %d. top/bot/mid was %d/%d/%d\n", i, top,bot,mid);
    } else {
      ti->hit++;
    }

    //Statistics
    if(dump_statistics && !(i&0x0FFF)){
      u64 us;
      gettimeofday(&tv[1], NULL);
      us = getTimeUS(&tv[1], &tv[0]);

      printf("[Thread %d] Running stats: Processed input %d. Hits: %d Misses: %d  Cached: %d   Avg: %.3lf us\n", 
	     ti->id, i,ti->hit, ti->miss, exact_hit,
	     ((double)us)/4096.0 );
      gettimeofday(&tv[0], NULL);
    }

  }

  fclose(fs);
  if(fu) fclose(fu);
  FREE(v1);
  FREE(v2);

  ti->counted = count;
  return;
}
//=============================================================================
void launch(char *fn_unsorted, char *fn_sorted,
	    int num_threads,  u64 vector_limit,
	    u64 tot_index_mem,
	    int table_form){

  u64 i,j;

  pthread_attr_t attr;
  pthread_t     *threads;
  thread_info_t *tdat;
  int status;

  index_t *index_block;

  u64 tot_hit, tot_miss;
  u64 us_init, us_proc, us_tot;

  //printf("Launching %d threads w/ %llu total_mem and vlimit of %llu\n", num_threads, tot_index_mem, vector_limit);

  if(clear_cache)
    clearSystemDiskCache();

  //Create an index to (hopefully?) speed things up. All threads share the same index (RO)
  timerMark(0); 
  createIndex(fn_sorted, tot_index_mem, &index_block);
  timerMark(1);


  //Setup the threads
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  tdat = (thread_info_t *)mallocOrDie(num_threads*sizeof(thread_info_t));
  threads = (pthread_t *)mallocOrDie(num_threads*sizeof(pthread_t));

  //Create and launch
  for(i=0; i<num_threads; i++){
    tdat[i].id                  = i;
    tdat[i].num_threads         = num_threads;
    tdat[i].total_input_vectors = vector_limit;
    tdat[i].fn_unsorted         = fn_unsorted;
    tdat[i].fn_sorted           = fn_sorted;
    tdat[i].counted             = NULL;
    tdat[i].hit                 = 0;
    tdat[i].miss                = 0;
    tdat[i].index_block         = index_block;

    pthread_create( &threads[i], &attr, th_cmp, (void *)&tdat[i]);
  }


  //Merge
  for(i=0;i<num_threads; i++)
    pthread_join(threads[i], (void **)&status);

  timerMark(2);

  //Do the general check
  for(j=0, tot_hit=0, tot_miss=0;j<num_threads;j++){
    tot_hit  += tdat[j].hit;
    tot_miss += tdat[j].miss;
  }


  //Print out the statistics
  us_init = timerMarkDiffUS(1,0);
  us_proc = timerMarkDiffUS(2,1);
  us_tot  = timerMarkDiffUS(2,0);


  if(table_form){
    //Tabbed form to make it easier to pull into spreadsheets
    printf("%d\t%llu\t%s\t"
	   "%llu\t%llu\t"
	   "%llu\t%llu\t"
	   "%.3lf\t%.3lf\n",
	   num_threads, tot_index_mem, fn_sorted,
	   tot_hit, tot_miss,
	   us_init/1000,
	   us_proc/(1000*1000),
	   (((double)us_proc)/ ((double)vector_limit)),
	   (((double)us_tot )/ ((double)vector_limit))  );

  } else {
    printf("Threads: %d  IndexBufsize: %llu  SortedFile: %s  Hit/Miss: %llu/%llu  InitTime: %llu ms  ProcessTime: %llu s  AvgProc/vec:  %.3lf us   AvgAll/vec:  %.3lf us\n",
	 num_threads, tot_index_mem, fn_sorted,
	 tot_hit, tot_miss,
	 us_init/1000,
	 us_proc/(1000*1000),
	 (((double)us_proc)/ ((double)vector_limit)),
	 (((double)us_tot )/ ((double)vector_limit)) );
  }

  

  //Free up memory
  FREE(tdat);
  FREE(threads);
  if(tot_index_mem){
    FREE(index_block->data);
    FREE(index_block);
  }

}

//=============================================================================
int main(int argc, char **argv){

  u64 i,ich, j;
  float *buf1;
    
  FILE *fs=NULL;
  FILE *fu=NULL;
  char *fn_sorted=NULL;
  char *fn_unsorted=NULL;
  

  u64  vlimit=0;
  int  display=0;

  u64 vlen1, vnum1, vtype1;
  u64 vlen2, vnum2, vtype2;
  char *count;
  float *v1,*v2;

  u64 misses, multiples;
  u64 us;
  u64 num_threads=1;
  u64 max_threads=0;
  u64 tt, bb;

  u64 tot_index_mem=0;
  u64 max_index_mem=0;

  int exit_now=0;
  int table_form=0;

  //Pick off the switches/one args
  while( (ich=getopt(argc, argv, "t:T:b:B:n:u:CvdqQXh"))!=EOF){
    switch(ich){
    case 't': num_threads     = atoiUnits(optarg); break;
    case 'T': max_threads     = atoiUnits(optarg); break;

    case 'b': tot_index_mem   = atoiUnits(optarg); break;
    case 'B': max_index_mem   = atoiUnits(optarg); break;

    case 'n': vlimit          = atoiUnits(optarg); break;

    case 'u': fn_unsorted     = strdup(optarg);    break;

    case 'q': table_form      = 1; break;
    case 'Q':
      printf("Threads\tIndexBufsize(B)\tSortedFile\tHits\tMisses\tInitTime(ms)\tProcessTime(s)\tAvgProc/vec(us)\tAvgAll/vec(us)\n");
      break;

    case 'C': clear_cache     = 0; break;
    case 'v': dump_statistics = 1; break;
    case 'd': debug           = 1; break;
    case 'X': exit_now = 1; break;
    case 'h':
    default:
      dumpHelp();
      exit(0);
    }
  }

  if(exit_now) exit(0);

  if(optind < argc){
    fn_sorted = strdup(argv[optind]);
  } else {
    printf("Lacked a sorted file to work with\n");
    dumpHelp();
    exit(0);
  }

  //Open the sorted file. Must have it, and must be float type
  fs = openFileWithMetaOrDie(fn_sorted,   "r", &vlen1, &vnum1, &vtype1);
  fclose(fs);
  if(vtype1!=VT_F32)
    emsg("Sorted file is type '%s', currently only support 'float'", getVTypeStr(vtype1) );
  

  if(num_threads>THREADS_MAX)     emsg("Number of threads is larger than hard limit of %d",THREADS_MAX);
  if(max_threads<num_threads)     max_threads = num_threads;
  if(tot_index_mem>INDEX_MEM_MAX) emsg("Index memory is larger than hard limit of %llu", INDEX_MEM_MAX);
  if(max_index_mem<tot_index_mem) max_index_mem = tot_index_mem;

     
 

  //Check unsorted file if we're given it
  if(fn_unsorted){
    fu = openFileWithMetaOrDie(fn_unsorted, "r", &vlen2, &vnum2, &vtype2);
    if(vtype2!=vtype1)
      emsg("Unsorted file was type '%s', currently only support 'float'", getVTypeStr(vtype1));
    fclose(fu);
    if(!vlimit) vlimit = vnum2;
    else vlimit = (vlimit<vnum2) ? vlimit : vnum2;
  } else {
    
    //No unsorted file. Verify we've got a number of input vectors to work with
    if(!vlimit)
      emsg("Need to specifiy either the number of input vectors (-n vnum) or give an unsorted input file");

  }
  


  for(tt=num_threads; tt<=max_threads; tt*=2){
    for(bb=tot_index_mem; bb<=max_index_mem; bb = (bb) ? bb*16 : 512*1024 ){
    	//printf("tt=%d bb=%d\n", tt, bb);
      launch(fn_unsorted, fn_sorted, tt, vlimit, bb, table_form);
    }
  }




}
