/**
 * Name:  block_test.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#define _GNU_SOURCE 1          //for O_DIRECT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>

#include <pthread.h>

#include "common.h"


const u64 min_block_size = (512);
const u64 max_block_size = (1*1024*1024);



//u64 total_bytes    = 1024*1024*1024;
u64 interval_size  = 1024*1024;//8*1024;    //for random: to provide complete independence between blocks
u64 num_spots;                       //for random: = total_bytes/interval_size
u64 blocks_per_thread;               //

u64 max_bytes      = (8*1024*1024); //0xFFFFFFFFFFFFFFFFULL;
u32 block_size     = 8*1024; 
u64 block_ops_per_thread = 1024;



int do_random=0;
int do_read  =1;

u64 *results_bytes_read; //thread-writable


typedef struct {
  u32   seed_id;
  char *mem_block; //Must be aligned to 512 byte boundary
  int   f;
  u64   init_offset;
  u64   tot_transferred;
  
} state_info_t;

//=============================================================================
void * workerThread(void *thread_args){

  state_info_t *si;
  int f;
  s64 rc;
  
  int id, seed;
  u64 xfer_size = 0;
  u64 tot_transferred = 0;
  char *block;
  u64   offset;
  int i;

  si = (state_info_t *)thread_args;

  id = seed = si->seed_id;
  f         = si->f;
  block     = si->mem_block;
  offset    = si->init_offset;
  

  //Do all of the iops that this thread requires
  for(i=0; i<block_ops_per_thread; i++){


    //Random- Need to jump to a random offset before we do anything. 
    if(do_random){
      //Skip to a random block 
      offset = (u64) (((double)num_spots) * ((double)rand_r(&seed) / ((double)RAND_MAX+1.0)));
      offset *= interval_size;
 
      if(offset+block_size > max_bytes)
	emsg("Offset too big: %llu (+%llu) for %llu byte file for id %d\n",
	     offset, offset+block_size, max_bytes, id);

      //Jump to the offset
      rc = lseek(f, offset, SEEK_SET);
      if(rc<0)
	emsg("lseek failed on offset %llu into %llu byte file\n", offset, max_bytes);
    } else {

      //Sequential: Check to see if this read would put us over the top. If yes, wrap around.
      //Note: offset is not the right word here. It's really refers to what the offset will
      //      be AFTER the transfer. We do this here to prevent an op from going out of bounds.
      offset += block_size;
      if(offset>=max_bytes){
	offset=block_size; //because offset reflects value after read
	rc = lseek(f, 0, SEEK_SET);
	if(rc<0)
	  emsg("lseek failed on offset %llu into %llu byte file\n", offset, max_bytes);
      }
    }



    //Do the actual reads..
    if(1) { //Simple version: ignore short ops. Errors out so no bad data passed back

      //Do the actual read/write
      if(do_read) xfer_size = read( f, (void *)block, block_size);
      else        xfer_size = write(f, (void *)block, block_size);
      if(xfer_size != block_size)
	emsg("Short transfer: asked for %llu bytes and received %llu bytes (previously transferred %llu bytes offset %llu  on op %d)", 
	     block_size, xfer_size, tot_transferred, offset, i);
      tot_transferred += xfer_size;

    } else { //Retry version: retry until we get all data. This may lock up

      u64 tmp_size   = block_size;
      u64 tmp_offset = 0;
      do{

	//Do the actual read/write
	if(do_read) xfer_size = read( f, (void *)&block[tmp_offset], tmp_size);
	else        xfer_size = write(f, (void *)&block[tmp_offset], tmp_size);

	tmp_size        -= xfer_size;
	tmp_offset      += xfer_size;
	tot_transferred += xfer_size;
	
	if(xfer_size!=block_size) printf("Transfer %llu bytes instead of %llu\n", xfer_size, block_size);
      } while(tmp_size);

    } 

    //{ u64 *x = (u64 *) block;
    //  printf("Offset is %llu, Val is %llu (bytes: %llu)\n", offset/block_size, x[0], x[0]*block_size);
    //}

    //printf("<%d> total %llu  bs=%llu offset=%llu num_spots=%llu\n",
    //	   id, tot_transferred, block_size, offset, num_spots);
  }

  if(!do_read) sync(); //Close out writes.. may or may not work right

  close(f);
  si->tot_transferred = tot_transferred;
  return 0;

}




//=============================================================================
void dumpHelp(void){
  printf("block_test <test type> <test options> file_name\n"
         " test type options\n"
         "   -r              : Read test\n"
         "   -w              : Write test\n"
         "   -S              : Sequentially access blocks\n"
         "   -R              : Randomly access blocks\n"
         "\n"
         "   -m              : Assign a unique file to each thread (filenameX, X=thread number)\n"
         "   -d              : Open files with O_DIRECT flag (bypass disk cache)\n"
         "   -v              : Print debug information as tests are running\n"
	 "\n"
         "   -b block_size   : Specify how many bytes to use in a file transaction\n"
         "   -t num_threads  : Specify how many threads to use (final performance result is aggregate)\n"
         "   -s max_bytes    : Put an upper bound on how much of the file is to be accessed [1]\n"
	 "   -i iterations   : how many block transfers to perform total [1k]\n"
	 "\n"
	 "   -q              : print results in tabular form\n"
	 "   -Q              : print tabular form header info and exit\n"
	 "\n"
	 "   file_name       : The file to work with (or the base name of the files if -m)\n"
	 "\n"
         "  Purpose: calculates amount of time required to randomly read through a file\n");
	 
  exit(0);
}


//=============================================================================
int main(int argc, char **argv){

  s64 rc;
  pthread_t *tid;
  state_info_t *si;

  int num_threads=4;
  int block_ops=1024;
  int multiple_files=0;
  int direct=0;

  char **filename_list;
  char *filename_base=NULL;
  u64  smallest_file_size;
  int file_flags;

  u64 us;
  u64 status;
  u64 tot_transferred;
  int table_form=0;
  int table_top_only=0;
  int i,ich;


  if(argc<2) dumpHelp();
  common_init();
 

  while( (ich=getopt(argc, argv, "rwSRmdvqQb:s:i:t:h"))!=EOF){
    switch(ich){
    //Types of test to perform
    case 'r': do_read=1;   break;
    case 'w': do_read=0;   break;
    case 'S': do_random=0; break;
    case 'R': do_random=1; break;

    case 'm': multiple_files=1; break;
    case 'd': direct=1; break;
    case 'v': debug=1; break;
    case 'b': block_size  = atoiUnits(optarg); break;
    case 's': max_bytes   = atoiUnits(optarg); break;
    case 't': num_threads = atoi(optarg);      break;
    case 'i': block_ops   = atoiUnits(optarg); break;

    case 'q': table_form=1;     break;
    case 'Q': table_top_only=1; break;

    case 'h':
    default:
      dumpHelp();
      exit(0);
    }
  }

  if(table_top_only){
    printf("Time(us)\tRandom\tRead\tDirect\tThreads\tBlockSize\tPerThreadSize\tAggBytesTransferred\tkIOPS\tMB/s\n");
    exit(0);
  }


  //Grab the input filename or die
  if(optind<argc)
    filename_base = strdup(argv[optind++]);
  else{
    printf("Need a file to work with\n");
    dumpHelp();
  }

  //Check block size
  if((block_size<min_block_size)||(block_size>max_block_size))
    emsg("Error: block size out of range (%llu-%llu)\n",min_block_size, max_block_size);


  if((num_threads<0) || (num_threads>16))
    emsg("Number of threads must be 1-16");



  //Check file size and create if necessary
  filename_list = (char **)malloc(num_threads*sizeof(char *));
  if(multiple_files){

    smallest_file_size = max_bytes;

    for(i=0; i<num_threads; i++){
      filename_list[i] = (char *)malloc(strlen(filename_base)+8);
      sprintf(filename_list[i],"%s%d", filename_base, i);

      rc = createMissingFileOfMinSize(filename_list[i], max_bytes,0);
      if(rc<0){
	//file exists, but isn't large enough.
	smallest_file_size = getFileBytes(filename_list[i]);
      }
    }

    if(smallest_file_size < max_bytes){
      warn("File %sX exists, but is smaller than the requested %llu size. Tests limited to %llu bytes",
	   filename_base, max_bytes, smallest_file_size);
      max_bytes = smallest_file_size;
    }

  } else {

    rc = createMissingFileOfMinSize(filename_base, max_bytes,1);

    if(rc<0){
      smallest_file_size = getFileBytes(filename_base);
      warn("File %s exists, but is smaller than the requested %llu size. Tests limited to %llu bytes",
	   filename_base, max_bytes, smallest_file_size);
      max_bytes = smallest_file_size;
    }
    //Same file for everyone
    for(i=0; i<num_threads; i++)
      filename_list[i] = filename_base;

  }

  //Slice up the file into blocks
  num_spots = (max_bytes/interval_size) - 2*(block_size/interval_size);

  dbg("num_spots=%llu max_bytes=%llu interval=%llu block_size=%llu", num_spots, max_bytes, interval_size);

  //Figure out how many operations we're going to do
  block_ops_per_thread = block_ops/num_threads;
  block_ops = block_ops_per_thread * num_threads; //to overcome rounding
  if(block_ops > num_spots)
    warn("The number of operations exceeds the number of blocks in the input. Expect caching effects.");
  if(block_size > interval_size)
    warn("The block size (%llu) is bigger than the interval size (%llu). Expect some caching effects.", block_size, interval_size);


  tid                = (pthread_t *)   malloc(num_threads * sizeof(pthread_t));
  results_bytes_read = (u64 *)         malloc(num_threads * sizeof(u64));
  si                 = (state_info_t *)malloc(num_threads * sizeof(state_info_t));


  //Set the flags for how we open the files
  file_flags =   ((do_read) ? O_RDONLY : O_WRONLY)  
	       | ((direct)  ? O_DIRECT : 0);


  dbg("Block size: %d num_threads=%d max_bytes=%llu Iterations=%llu",
      block_size, num_threads, max_bytes, block_ops);

  //Create all the high-weight thread state in advance
  for(i=0; i<num_threads; i++){
    char *block_unaligned;
    char *block;

    //Direct requires aligned memory
    block_unaligned = (char *)malloc(block_size+512);
    block           = (char *)(((unsigned long)block_unaligned + 512-1) & (~(511)));

    si[i].seed_id   = i;
    si[i].mem_block = block;
    si[i].f         = open(filename_list[i], file_flags);
    if(si[i].f<0)
      emsg("Couldn't open %s", filename_list[i]);

    //Try to prevent caching by starting threads at different points in
    //the file.
    if(do_random || multiple_files){
      //Threads have some independence: Just start from beginning of file
      si[i].init_offset=0;
    } else {
      //Sequential read of one file: Start at different points in file to prevent caching
      si[i].init_offset = interval_size * i * (num_spots/num_threads);
    } 

  }


  //-- Launch Threads --
  timerMark(0); 
  for(i=0;i<num_threads;i++){
    rc = pthread_create( &tid[i], NULL, workerThread, (void *)&(si[i]));
    if(rc)
      emsg("Error creating thread %d",i);
  }

  dbg("Launched %d threads", num_threads);

  
  //-- Wait for all to be done --
  for(i=0; i<num_threads;i++){
    rc = pthread_join( tid[i], (void **)&status);
    if(rc)
      emsg("Join error");
  }
  timerMark(1);  
  us = timerMarkDiffUS(1,0);
  

  //-- Sum up the results --
  tot_transferred = 0;
  for(i=0;i<num_threads;i++){
    tot_transferred += si[i].tot_transferred;
    dbg("Thread %d count: %llu", i, si[i].tot_transferred); 
  }

 
  

  if(table_form){

    //This gives tab separated values that are easier to pull into spreadsheets 
    printf("%llu\t"
	   "%d\t%d\t%d\t"
	   "%d\t%d\t%llu\t"
	   "%llu\t%.3lf\t%.3lf\n",
	   us, 
	   do_random, do_read, direct,
	   num_threads, block_size, tot_transferred/num_threads,
	   tot_transferred, 1000.0*((double)block_ops)/((double)us), ((double)tot_transferred)/((double)us));

  } else {

    //This gives a more human readable output
    printf("Time: %llu us %s %s %s Threads: %d BlockSize: %d PerThreadSize: %llu  AggBytesTransferred: %llu  Iops: %.3lf MIOPs  MB/s: %.3lf\n", 
	   us, 
	   (do_random) ? "Random" : "Sequential",
	   (do_read)   ? "Reads"  : "Writes",
	   (direct)    ? "(Direct-Mode)" : "(Non-direct)",
	   num_threads,
	   block_size,
	   tot_transferred/num_threads, 
	   tot_transferred, 
	   ((double)block_ops)/((double)us),
	   ((double)tot_transferred)/((double)us));
  }

  return 0;
}
