/**
 * Name:  common.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <float.h>
#include <pthread.h>

#include <fcntl.h>
#include <sys/mman.h>

#include <assert.h>
#include "common.h"


struct timeval *_tv=NULL;


int debug=0;

//=============================================================================
void common_init(void){
  _tv  = (struct timeval *) malloc(MAX_TIMERS*sizeof(struct timeval));
}
//=============================================================================
u64 getTimeUS(struct timeval *stop, struct timeval *start){
  return (((u64)(stop->tv_sec)  - (u64)(start->tv_sec))*1000000) + 
          ((u64)(stop->tv_usec) - (u64)(start->tv_usec));
}
//=============================================================================
void timerMark(int spot){
  if(!_tv) common_init();
  assert((spot>=0)&&(spot<MAX_TIMERS));
  gettimeofday(&_tv[spot], NULL);
}
//=============================================================================
u64 timerMarkDiffUS(int stop_spot, int start_spot){

  if(!_tv) {emsg("No timer allocated?");}
  assert((stop_spot >=0)&&(stop_spot <MAX_TIMERS));
  assert((start_spot>=0)&&(start_spot<MAX_TIMERS));
  return getTimeUS(&_tv[stop_spot],&_tv[start_spot]);
}

//=============================================================================
u64 getFileBytes(char *filename){
  int rc;
  static struct stat fstat_buf;
  rc = stat(filename, &fstat_buf);
  if(rc) { emsg("Got bad stat for %s", filename); }
  return fstat_buf.st_size;
}

//=============================================================================
int  fileExists(char *filename){
  int rc;
  static struct stat fstat_buf;
  return !stat(filename, &fstat_buf);  
}

//=============================================================================
// Returns: -1 = file already exists, but is smaller than min_size
//           1 = file already exists and is large enough
//           2 = file created and filled out
//         
int createMissingFileOfMinSize(char *filename, u64 min_size, int doit_fast){
  
  FILE *f;
  int rc;
  const int bundles=1024;  //set high enough to do many at a time
  const int bbytes=4*1024; //set low enough to tag pages
  u64 i,j;
  u64 *buf;
  u64 left;
  u64 cur;


  //Check to see if file exists already and is big enough
  if(fileExists(filename)) {
    u64 orig_size;
    orig_size = getFileBytes(filename);
    if(orig_size<min_size) return -1; //not big enough
    return 1; //big enough
  }

  //File needs to be created and filled
  f = fopen(filename, "w");
  if(!f)
    emsg("Could not create file %s of size %llu", filename, min_size);


  //Write data to make this a big file
  if(doit_fast){
    //Just tag the last byte in the file. This may not be good for benchmarks 
    //because all the inbetween pages may be fake-cached.
    rc = fseeko(f,min_size-1,SEEK_SET);
    if(rc<0)
      emsg("Failed to seek when creating file %s of size %llu\n", filename, min_size);

    rc = 0;
    rc = fwrite(&rc, 1, 1, f);
    if(rc!=1)
      emsg("Failed to write when creating file %s of size %llu\n", filename, min_size);
    
  } else {
    //actually write a bunch of junk out.

    //Do several bundles at a time
    buf = (u64 *)malloc(bundles*bbytes);
    if(!buf) {emsg("Couldn't malloc");}
    for(i=0, left=min_size, cur=bundles*bbytes; left>0; i++,left-=cur){
    
      //Tag each bundle with a different marker for sanity
      for(j=0;j<bundles;j++)
	buf[j*bbytes/sizeof(u64)] = i*bundles + j;
    
      if(left<cur) cur=left;
    
      rc = fwrite(buf, cur, 1, f);
      if(rc!=1)
	emsg("Failed to write when creating file %s of size %llu\n", filename, min_size);
    
    }
    free(buf);
    fflush(f);
    fclose(f);
  }

  return 2; //File created and filled out
}

//=============================================================================
FILE * safeOpenFile(char *path, char *name, char *direction){

  char tname[256];
  FILE *f;

  if((!name) || (!direction))
    emsg("No file name/direction supplied to safeOpenFile");

  if(path) sprintf(tname,"%s/%s", path,name);
  else     strcpy(tname, name);

  if(!strcmp(tname,"STDOUT"))
    if(direction[0]=='w')
      return stdout;

  if(!strcmp(tname,"STDIN"))
    if(direction[0]=='r')
      return stdin;


  f = fopen(tname, direction);
  if(!f)
    emsg("Couldn't open file %s",tname);

  return f;
}

int _read_files=0;
int _write_files=0;

//=============================================================================
//This opens a file and gets/puts meta information in its header
FILE *openFileWithMeta(char *name, char *direction,
		       u64 *vector_length, u64 *vector_num, u64 *vector_type ){

  FILE *f;
  u64 *config;
  s64 rc;

  if(!name || !direction)
    emsg("Null name or direction given");

  f = fopen(name, direction); 
  if(!f) return f;

  if(direction[0]=='w') _write_files++;
  if(direction[0]=='r') _read_files++;

  config = (u64 *)calloc(file_config_words, sizeof(u64));
  if(!config)
    emsg("malloc error");

  switch(direction[0]){
  case 'r':
  case 'R':
    //Read in configuration data
    rc = fread(config, sizeof(u64), 3, f);
    if(rc!=3)
      emsg("Couldn't read all configuration data from '%s'",name);

    if(vector_length) *vector_length = config[0];
    if(vector_num)    *vector_num    = config[1];
    if(vector_type)   *vector_type   = config[2];

    //Skip rest ofconfiguration space
    rc = fseeko(f, file_config_bytes, SEEK_SET);
    if(rc)
      emsg("Couldn't seek to right location in '%s'", name);

    break;

  case 'w':
  case 'W':
    //Write out configuration data
    if((!vector_length) || (!vector_num) || (!vector_type))
      emsg("Null vector length/num/type pointer in openfilewithmeta");

    config[0] = *vector_length;
    config[1] = *vector_num;
    config[2] = *vector_type;

    rc = fwrite(config, sizeof(u64), file_config_words, f);
    if(rc!=file_config_words)
      emsg("Problem writing configuration data to '%s'", name);
    break;

  default:
    emsg("Unknown open file direction '%s'", direction);
  }


  free(config);
  return f;
}
//=============================================================================
FILE * openFileWithMetaOrDie(char *name, char *direction,
			     u64 *vector_length, u64 *vector_num, u64 *vector_type ){

  FILE *f;
  f = openFileWithMeta(name, direction, vector_length, vector_num, vector_type);
  if(!f){
    perror(" emsg ");
    emsg("Faile to open '%s' with perms '%s'. NumOpened (r/w): %d/%d",
	   name, direction, _read_files, _write_files);
    exit(1);
  }

  return f;
}
//=============================================================================
void createReducedFile(char *dst_name, char *src_name, 
		       double fraction, int select_random_locations){

  FILE *fi, *fo;
  u64 vlen, vnum, vtype;
  size_t vlen_bytes;
  u64 vnum_new;
  u64 vskip;
  u64 i;
  void *buf;
  int rc;

  fi = openFileWithMetaOrDie(src_name, "r", &vlen, &vnum, &vtype);


  vnum_new   = (u64) ((double)vnum * fraction);
  vskip      = vnum/vnum_new;
  vlen_bytes = vlen * getVTypeSize(vtype);

  
  if(vnum_new>vnum) emsg("got bad number of vectors");
  if(vnum_new<1)    vnum_new=1; //always make something


  fo = openFileWithMetaOrDie(dst_name, "w", &vlen, &vnum_new, &vtype);
  buf = (void *) mallocOrDie(vlen_bytes);

  for(i=0; i<vnum_new; i++){

    if(!select_random_locations){
      //Skip a few lines. Note- vskip shouldn't result in exceeding input length
      //because of div.  eg 100/19 = 5, 5*19<100
      if(i){ //first time doesn't advance, and don't count current read in batch
	rc = fseeko(fi, (vskip-1)*vlen_bytes, SEEK_CUR);
	if(rc) emsg("Bad seek?");
      } 
    } else {
      //Just select a random vector in the file
      vskip = (u64) (((double) vnum) * (rand()/(RAND_MAX+1.0)));
      rc = fseeko(fi, (file_config_words * sizeof(u64)) +  vskip*vlen_bytes, SEEK_SET);
      if(rc) emsg("bad seek?");
    }
    
    rc = fread(buf, vlen_bytes, 1, fi);
    if(rc!=1) emsg("Bad read: %d entries",rc);

    rc = fwrite(buf, vlen_bytes, 1, fo);
    if(rc!=1) emsg("Bad write");
  }

  free(buf);
  fclose(fi);
  fclose(fo);

}

//=============================================================================
void seedRand(){
  srand(time(0));
}

void generateRandVector(void *vec, int vector_length, u64 vector_type){
  int i;

  for(i=0;i<vector_length; i++){
    switch(vector_type){
    case VT_U32: ((u32 *)vec)[i]    =  (u32)rand(); break; //possibly bad
    case VT_U64: ((u64 *)vec)[i]    =  (u64)rand() || (((u64)rand())<<32); break; //possibly bad
    case VT_F32: ((float *)vec)[i]  = ((float)rand()) / ((float)RAND_MAX);   break; //0-1.0
    case VT_F64: ((double *)vec)[i] = ((double)rand()) / ((double)RAND_MAX); break; //0-1.0
    default:
      emsg("Unknown type");
    }
  }
}

//=============================================================================
char *getVTypeStr(u64 x){
  switch(x){
  case VT_U32 : return "U32";
  case VT_U64 : return "U64";
  case VT_F32 : return "float";
  case VT_F64 : return "double";
  default:      return "Unknown";
  }
}
//=============================================================================
u64 getVTypeFromStr(char *x){

  if(!strcasecmp(x,"u32"))    return VT_U32;
  if(!strcasecmp(x,"u64"))    return VT_U64;
  if(!strcasecmp(x,"float"))  return VT_F32;
  if(!strcasecmp(x,"double")) return VT_F64;

  warn("Unknown string type '%s'", x);
  return 0;
}
//=============================================================================
size_t getVTypeSize(u64 t){
  switch(t){
  case VT_U32 : return sizeof(u32);
  case VT_U64 : return sizeof(u64);
  case VT_F32 : return sizeof(float);
  case VT_F64 : return sizeof(double);
  default: 
    emsg("Unknown type (%llu)", t);
  } 
  return 0;
}

//=============================================================================
void dumpVector(void *x, u64 vlen, u64 vtype, u64 hlimit){
  int j;

  //Dump the columns we're interested in
  for(j=0; (j<vlen) && !((hlimit) && (j>hlimit)); j++){
    switch(vtype){
    case VT_U64 : printf(" %llu",  ((u64 *)x)[j]); break;
    case VT_F32 : printf(" %.3f",  ((float *)x)[j]); break;
    case VT_F64 : printf(" %.3lf", ((double *)x)[j]); break;
    default: emsg("Unknown type");
    }
  }
  printf("\n");     
}
//=============================================================================
void dumpFile(char *fname, u64 hlimit, u64 vlimit){

  u64 i,j, vlen, vnum, vtype;
  
  FILE *f;
  char *x;
  int rc;

  f = openFileWithMetaOrDie(fname, "r", &vlen, &vnum, &vtype);

  x = (char *) malloc(vlen * sizeof(u64)); //pick largest


  printf("File '%s': VecLen=%llu VecNum=%llu VecType=%s\n",
	 fname, vlen, vnum, getVTypeStr(vtype));

  for(i=0; (i<vnum) && !((vlimit) && (i>vlimit)); i++){
    printf("[%4llu]: ",i);

    //Grab next vector
    rc = fread(x, (vtype==VT_F32) ? sizeof(float) : sizeof(double), vlen, f);
    if(rc!=vlen)
      emsg("Bad read on file '%s' (rcl=%d)", fname,rc);
    
    dumpVector((void *)x, vlen, vtype, hlimit);

  }

  FREE(x);
  fclose(f);
}

//=============================================================================
void * mallocOrDie(size_t size){
  void *res;
  res = malloc(size);
  if(!res){
    emsg("Could not malloc %d bytes", size);
  }
  return res;
}

//=============================================================================
void * callocOrDie(size_t nmemb, size_t size){
  void *res;
  res = calloc(nmemb, size);
  if(!res){
    emsg("Could not calloc %d bytes", size);
  }
  return res;
}

//=============================================================================
// Simple addition to atoi so use can specify kilo,mega,giga
u64 atoiUnits(char *s){
  
  u64 val;

  val = atoi(s);

  switch(s[strlen(s)-1]){
  case 'k':
  case 'K':
    val *= 1024;
    break;  
  case 'm':
  case 'M':
    val *= 1024*1024;
    break;
  case 'g':
  case 'G':
    val *= 1024*1024*1024;
    break;
  default:
    break;
  }

  return val;
}

//=============================================================================
int clearSystemDiskCache(){

  uid_t uid;
  static int have_warned = 0;
  int rc;

  //First: use drop_caches if we have root access.
  uid = geteuid();
  if(uid==0){
    FILE  *f;

    sync(); //Probably does nothing.
 
    //Note: drop caches is destructive for uncommited writes!
    f = fopen("/proc/sys/vm/drop_caches","w");
    if(!f) {
      warn("Problem opening up disk cache");
      return -2;
    }

    //Values have different meanings, according to http://linux-mm.org/Drop_Caches
    // 1 : FREE pagecache
    // 2 : FREE dentries and inodes
    // 3 : FREE pagecache, dentries, and inodes
    fprintf(f, "3\n");
    fclose(f);
  } else {
    //No root access, pass a warning..
    if(!have_warned)
      warn("Didn't have root access, so couldn't clear cache with guarantees");
    have_warned=1;
  }

  //Create a big file if we haven't done so yet
  rc=createMissingFileOfMinSize("/tmp/bigfile.dat", 0x08LLU<<30, 1);
  if(rc<1){
    warn("Could not create /tmp/bigfile.dat for thwarting cache");
    return 1;
  }

  //Use DD to copy it
  rc=system("dd if=/tmp/bigfile.dat of=/dev/null >& /dev/null");

  return rc;
}
//=============================================================================
// th_fileStreamer: a thread that simply reads from a vector file sequentially
//                  and streams blocks of data through a queue to the user
void * th_fileStreamer(void *arg){
  
  volatile file_streamer_t *x;
  FILE *f;
  u64  tmp_words, vecs_left, tmp_type; //For info about file
  u64  cur_num;
  u64  spot, spot_p1;
  u64  vec_bytes;
  u64  total_bytes;
  u64  cur;
  int  rc;
  struct timeval tv[2];
  u64  us;

  u64  dbg_num_reads=0;
  u64  dbg_bytes_read=0;


  x = (file_streamer_t *)arg;

  f = openFileWithMetaOrDie(x->filename, "r", &tmp_words, &vecs_left, &tmp_type);
  

  //Figure out how big a slot is
  vec_bytes        = tmp_words*getVTypeSize(tmp_type);
  total_bytes      = vec_bytes*vecs_left;
  x->slot_bytes    = x->max_buf_bytes / x->slot_num;
  x->vecs_per_slot = x->slot_bytes / vec_bytes;
  x->slot_bytes    = x->vecs_per_slot * vec_bytes;
  x->done = 0;
  x->MBps = 0.0;

  

  if(x->slot_num<=2)
    emsg("FileStreamer must have at least two slots");

  if(x->slot_bytes < vec_bytes)
    emsg("Slot size (%llu) not large enough for a single vector (%llu)", x->slot_bytes, vec_bytes);


  //Allocate space
  x->buf = (char *)malloc(x->slot_num * x->slot_bytes);
  x->vecs_in_slot = (u64 *)malloc(x->slot_num * sizeof(u64));
  if((!x->buf) || (!x->vecs_in_slot))
    emsg("Malloc problem");

  gettimeofday(&tv[0], NULL);

  x->slot_front = x->slot_back = spot = 0;
  spot_p1 = 1;
  cur = x->vecs_per_slot;

  //Main Reading Loop
  while(vecs_left){

    //Make sure slot is available
    while( spot_p1 == (x->slot_front % x->slot_num)) sched_yield(); 

    //Adjust length if last read
    if(vecs_left < cur) cur=vecs_left;

    //Grab the data
    rc = fread(&x->buf[ spot*x->slot_bytes], vec_bytes, cur,f);
    if(rc!=cur)
      emsg("Failed to read %llu vectors. Got %d instead", cur, rc);

    dbg("Loaded slot %llu with %llu vects. Back=%llu Front=%llu", spot, cur, x->slot_back+1, x->slot_front);

    x->vecs_in_slot[spot] = cur;

    spot = spot_p1;
    spot_p1 = (spot_p1 + 1) % x->slot_num;

    vecs_left -= cur;
    x->slot_back++; //triggers update

    dbg_bytes_read += (vec_bytes*rc);
    dbg_num_reads++;
  }
  
  gettimeofday(&tv[1], NULL);

  fclose(f);

  us = getTimeUS(&tv[1],&tv[0]);
  x->MBps = ((double)total_bytes)/((double) us);

  printf("Thread done. Bytes read: %llu MiB  Num Reads: %llu  Perf: %.3lf MB/s  total_bytes: %llu time: %llu\n", 
	 dbg_bytes_read/(1000*1000), dbg_num_reads, x->MBps, total_bytes, us);
  fflush(NULL);

  x->done=1;

  //Memory freed later. Want to exit as fast as possible to let other threads work
  pthread_exit(NULL);
  return NULL;
}
//=============================================================================
int fileStreamer_launch(char *filename, 
			u64 max_buf_bytes,
			u64 num_slots,
			file_streamer_t *x){

  pthread_t tid;

  assert(x);

  memset(x,0,sizeof(file_streamer_t));
  x->filename      = strdup(filename);
  x->max_buf_bytes = max_buf_bytes;
  x->slot_num      = num_slots;
  
  return pthread_create( &tid, NULL, th_fileStreamer, x);
  
}
//=============================================================================
int fileStreamer_getNextPtr(volatile file_streamer_t *x,
			    void **ptr, u64 *num_vecs){


  u64 front;

  //Free the previously used slot 
  if(x->not_first_time){
    x->slot_front++;
  }

  assert((x->slot_back - x->slot_front) < x->slot_num);

  while(x->slot_front==x->slot_back){
    if(x->done){
      if(x->buf)          free(x->buf);
      if(x->vecs_in_slot) free(x->vecs_in_slot);
      x->buf = NULL;
      x->vecs_in_slot = NULL;
      dbg("Extract: Done detected\n");
      return -1; //Completed
    }
    
    sched_yield();
  }


  //Grab the slot and number of vals in slot
  front = x->slot_front % x->slot_num;
  *ptr = (void *) &(x->buf[ front * x->slot_bytes]);
  *num_vecs = (u64) x->vecs_in_slot[front];

  dbg("ReadST slot %llu with %llu vects Back=%llu Front=%llu", front, *num_vecs, x->slot_back, x->slot_front);
  
  x->not_first_time=1;

  return 0; 
}
//=============================================================================
//Borrowed from c.happycodings.com
int copyFile(char *dst, char *src){

  int i,o;
  size_t file_size;
  void *src_mem, *dst_mem;

  if((i=open(src, O_RDONLY))==-1)
    return -1;

  if((o=open(dst, O_RDWR|O_CREAT|O_TRUNC, 0666))==-1)
    return -2;

  file_size = lseek(i,0,SEEK_END);
  lseek(o, file_size-1, SEEK_SET);
  write(o, '\0',1);

  if((src_mem = mmap(0,file_size,PROT_READ, MAP_SHARED, i, 0)) == (void *)-1)
    return -3;

  if((dst_mem = mmap(0,file_size,PROT_WRITE, MAP_SHARED, o, 0)) == (void *)-1)
    return -4;

  memcpy(dst_mem, src_mem, file_size);

  munmap(src_mem, file_size);
  munmap(dst_mem, file_size);

  close(i);
  close(o);

  return 0;

}
void copyFileOrDie(char *dst, char *src){

  int rc;

  rc = copyFile(dst,src);
  if(!rc) return;

  switch (rc){
  case -1: emsg("Problem opening source '%s'",src); break;
  case -2: emsg("Problem opening destination '%s'",dst); break;
  case -3: emsg("Problem mmaping source '%s'",src); break;
  case -4: emsg("Problem mmaping destination '%s'",dst); break;
  default:
    emsg("Unknown error?");
  }

}
