/**
 * Name:  common.h
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef COMMON_H
#define COMMON_H

#include <sys/time.h>


#define MAX_TIMERS (256)

#define emsg(a, args...) {fprintf(stderr,"Error: " a "\n", ## args); exit(-1); }
#define info(a, args...) { if(debug>0) printf("DBG:" a "\n", ## args); }
#define dbg(a, args...) { if(debug>1) printf("DBG:" a "\n", ## args); }

//#define dbg(a, args...)
#define warn(a, args...) { fprintf(stderr, "Warning:" a "\n", ## args); }
#define FREE(a) { if(!a) emsg("Free on NULL pointer?"); free(a); a=NULL; }


//Vector data type in files
#define VT_U32  (1)  //Unsigned 32b int
#define VT_U64  (2)  //Unsigned 64b int
#define VT_F32  (4)  //Float  (32b)
#define VT_F64  (8)  //Double (64b)


typedef unsigned long long u64;
typedef unsigned int       u32;
typedef          long long s64;

#define file_config_bytes  (4*1024)
#define file_config_words  (4*1024/sizeof(u64)) //pad to make data aligned


extern int debug;




typedef struct {
  char *filename;
  u64   max_buf_bytes; //(Input):Max amount of space we can buffer
  
  char *buf;
  u64   slot_front;
  u64   slot_back;
  u64   slot_num;      //(Input):Number of slots to divide buffer into
  u64   slot_bytes;
  u64   vecs_per_slot;
  u64  *vecs_in_slot;  //Array saying how many vecs are in each slot. For short read on last
  int    done;
  int    not_first_time; //For reader
  double MBps; //Performance: how long it took to read all data
} file_streamer_t;




//Simple timing functions
u64    getTimeUS(struct timeval *stop, struct timeval *start);
void   timerMark(int spot);
u64    timerMarkDiffUS(int stop_spot, int start_spot);



//Basic file i/o
u64    getFileBytes(char *filename);  //figure out # bytes in a file or die
int    fileExists(char *filename);   //see if file exists
int    createMissingFileOfMinSize(char *filename, u64 min_size, int doit_fast);
FILE * safeOpenFile(char *path, char *name, char *direction);
FILE * openFileWithMeta(char *name, char *direction, u64 *vector_length, u64 *vector_num, u64 *vector_type );
FILE * openFileWithMetaOrDie(char *name, char *direction, u64 *vector_length, u64 *vector_num, u64 *vector_type );
void   createReducedFile(char *dst_name, char *src_name, double fraction, int select_random_locations);
void * mallocOrDie(size_t size);
void * callocOrDie(size_t nmemb, size_t size);

char * getVTypeStr(u64 x);
u64    getTypeFromStr(char *x);
size_t getVTypeSize(u64 t);

void   dumpVector(void *x, u64 vlen, u64 vtype, u64 hlimit);
void   dumpFile(char *fname, u64 hlimit, u64 vlimit);


u64    atoiUnits(char *s);

//Random number stuff
void seedRand();
void generateRandVector(void *vec, int vec_len, u64 vec_type);


int clearSystemDiskCache();

void * th_fileStreamer(void *arg);
int fileStreamer_launch(char *filename, u64 max_buf_bytes, u64 num_slots, file_streamer_t *x);
int fileStreamer_getNextPtr(volatile file_streamer_t *x, void **ptr, u64 *num_vecs);


int copyFile(char *dst, char *src);
void copyFileOrDie(char *dst, char *src);

#endif

