/**
 * Name:  kmeans.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <float.h>

#include <pthread.h>
#include <assert.h>


#include "common.h"



typedef struct threadState_s {
  FILE   *f;            //The file we're working with
  char   *file_name;    //File to load
  int     thid;         //Our thread id (mt code only)
  u64     start_spot;   //id of vector to start with
  u64     num_spots;    //How many vectors to check
  int     k;            //How many vectors we're tracking
  int     vlen;         //How long the vectors are
  float **centerVectors;//Vector that we want to find a match for
  float **runningSums;  //each k-vector has running sum
  u64    *num_members;  //How many inputs are closest to each center
  float  *current_dist; //How far current center is from all its points (ie error)
  float  *input_vect;   //Space for next input vector we pull in
} threadState_t;




//=============================================================================
int allocThreadState(threadState_t *t, int thid,
		     int k, int vlen,
		     char *file_name,
		     u64 start_spot, u64 num_spots){
  int i;
  
  if(!t)
    emsg("Bad input?");

  dbg("Alloc[%d] k=%d vlen=%d start=%llu num=%llu", thid, k, vlen, start_spot, num_spots);

  t->f                = NULL;
  t->file_name        = strdup(file_name);
  t->thid             = thid;
  t->start_spot       = start_spot;
  t->num_spots        = num_spots;
  t->k                = k;
  t->vlen             = vlen;
  t->centerVectors    = (float **)mallocOrDie(k * sizeof(float *));
  t->runningSums      = (float **)mallocOrDie(k * sizeof(float *));
  t->num_members      =    (u64 *)mallocOrDie(k * sizeof(u64));
  t->current_dist     =  (float *)mallocOrDie(k * sizeof(float *));
  t->input_vect       =  (float *)mallocOrDie(vlen*sizeof(float));

  for(i=0;i<k;i++){
    t->centerVectors[i] = (float *)mallocOrDie(vlen*sizeof(float));
    t->runningSums[i]   = (float *)mallocOrDie(vlen*sizeof(float));
  }

  return 0;
}
//=============================================================================
int resetThreadStates(threadState_t *t, int num_threads){
  
  int i,j;  
  int rc;
  assert(t);

  for( i=0; i < num_threads; i++, t++){

    //Reopen the file
    if(t[i].f) fclose(t[i].f);
    t[i].f = openFileWithMetaOrDie(t[i].file_name, "r", NULL, NULL, NULL);
  
    //Move to this thread's offset
    rc = fseeko(t[i].f, t[i].start_spot * t[i].vlen * sizeof(float), SEEK_CUR);
    if(rc)
      emsg("Couldn't seek to initial position");

    //Zero out the statistics
    for(j=0; j<t[i].k; j++)
      memset(t[i].runningSums[j], 0, t[i].vlen * sizeof(float));

    memset(t[i].num_members,  0, t[i].k * sizeof(u64));
    memset(t[i].current_dist, 0, t[i].k * sizeof(float));
  }

  return 0;
}
//=============================================================================
int findNearestCenter(threadState_t *t, float *input_vector, 
		      int *center_id,   float *center_dist){

  float best_sum = FLT_MAX;
  int   best_id = -1;
  int i,j;
  float sum;
  float tmp;

  for(i=0; i < t->k; i++){
    
    for(sum=0.0 , j=0; (j < t->vlen) && (sum<best_sum); j++){
      tmp = (input_vector[j] - t->centerVectors[i][j]);
      sum += tmp*tmp;
    }
    if(j==t->vlen){ //new best bal
      best_id = i;
      best_sum = sum;
    }
  }

  if(center_id)   *center_id   = best_id;
  if(center_dist) *center_dist = best_sum;

  assert(best_id>=0);

  return 0;
}
//=============================================================================
void mergeAll(threadState_t *ts, int num_threads){
  
  int i,j,m;

  assert(ts);
  assert(num_threads);

  //Merge all results in first unit
  for(i=1; i < num_threads; i++){
    for(j=0; j < ts[0].k; j++){
      ts[0].num_members[j]  += ts[i].num_members[j];
      ts[0].current_dist[j] += ts[i].current_dist[j]; //tally error
      for(m=0; m < ts[0].vlen; m++)
	ts[0].runningSums[j][m] += ts[i].runningSums[j][m];
    }
  }
  //Average all the values
  for(j=0; j < ts[0].k; j++){

    //Average this vector's components in unit 0
    for(m=0; m < ts[0].vlen; m++)
      if(ts[0].num_members[j])
	ts[0].centerVectors[j][m] = ts[0].runningSums[j][m] / (float) ts[0].num_members[j];
   
    //Copy to other units
    for(i=1; i<num_threads; i++){
      memcpy(ts[i].centerVectors[j],
	     ts[0].centerVectors[j],
	     ts[0].vlen * sizeof(float));
    }
  }
}

//=============================================================================
void * th_do_kmeans(void *arg){

  u64 i;
  int j;
  threadState_t *ts;
  float *v;
  size_t rc;

  int best_id;
  float best_dist;

  ts = (threadState_t *)arg;  

  for(i=0; i < ts->num_spots; i++){

    //Grab next vector
    rc = fread(ts->input_vect, ts->vlen*sizeof(float), 1, ts->f);
    if(rc!=1)
      emsg("Bad read");

    //Find the center it's closest too
    findNearestCenter(ts, ts->input_vect, &best_id, &best_dist);

    //Update the statistics
    ts->current_dist[best_id] += best_dist;
    ts->num_members[best_id]++;

    for(j=0; j < ts->vlen; j++)
      ts->runningSums[best_id][j] += ts->input_vect[j];

  }
  return NULL;
}


//=============================================================================
void doWork(threadState_t *all_ts, int num_threads){
  
  int i;
  pthread_t *threads;
  pthread_attr_t attr;
  int status;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  threads = (pthread_t *)mallocOrDie(num_threads * sizeof(pthread_t));

  for(i=0; i<num_threads; i++){
    pthread_create( &threads[i], &attr, th_do_kmeans, (void *)&all_ts[i]);
  }

  for(i=0; i<num_threads; i++){
    pthread_join(threads[i], (void **)&status);
  }


}
//=============================================================================
void dumpHelp(void){
  printf("kmeans <-k #> <-i #> <-t #> filename\n"
         "    -k number of nearest neighbors to find\n"
	 "    -i number of iterations\n"
	 "    -t number of threads to use\n");
  exit(0);
}

//=============================================================================
int main(int argc, char **argv){

  int i,j,ich;
  int k=5;        //How many clusters to look for
  int iterations; //How many iterations to try
  u64 vlen=0;     //How long vectors are
  u64 vnum=0;     //How many input vectors are in file
  u64 vtype=0;    //What format the data is in

  char *filename=NULL;
  FILE *f;
  u64 config[2];
  int num_threads=1;
  u64 us;
  threadState_t *ts;
  u64 vec_per_thread;

  int table_top_only=0;
  int table_form=0;

  //Pick off the switches/one args
  while( (ich=getopt(argc, argv, "k:i:t:vVqQh"))!=EOF){
    switch(ich){
    case 'k': k           = atoiUnits(optarg); break;
    case 'i': iterations  = atoiUnits(optarg); break;     
    case 't': num_threads = atoiUnits(optarg); break;
    case 'v': debug       = 1; break;
    case 'V': debug       = 2; break;

    case 'q': table_form  = 1; break;
    case 'Q': table_top_only = 1; break;
    case 'h':
    default:
      dumpHelp();
      exit(0);
    }
  }

  if(table_top_only){
    printf("\n");
    exit(0);
  }
  
  //Anything else tagged on at end
  if(optind < argc){
    filename=strdup(argv[optind]);
  } 

  if((!filename)||(k<1)||(iterations<1)||(num_threads<1)){
    printf("Bad inputs..\n");
    dumpHelp();
  }


  f = openFileWithMetaOrDie(filename, "r", &vlen, &vnum, &vtype);

  info("Running k=%d iterations=%d threads=%d  vector_components=%llu vector_num=%llu",
       k, iterations, num_threads,
       vlen, vnum);


  //Allocate all the threads info
  vec_per_thread = vlen/num_threads;
  ts = (threadState_t *)mallocOrDie(num_threads*sizeof(threadState_t));
  for(i=0; i<num_threads; i++){            
    allocThreadState( &ts[i], i, k, vlen, filename, 
		      i*vec_per_thread,
		      (i<num_threads-1) ? vec_per_thread 
		                        : vnum-(i*vec_per_thread)); //last one gets extra    
  }
  

  //Generate random starting points
  seedRand();
  for(j=0; j<k; j++){
    generateRandVector(ts[0].centerVectors[j], vlen, VT_F32);
    dumpVector(ts[0].centerVectors[j], vlen, VT_F32, 3);
  }

  //Iterations
  for(i=0; i<iterations; i++){

    resetThreadStates(ts, num_threads);

    //Do work 
    doWork(ts, num_threads);

    //Combine the results
    mergeAll(ts, num_threads);

    //Dump Results
    if(debug){
      printf("Iteration %3d: Dist/Num:",i);
      for(j=0;j<k; j++)
	printf(" %.3f/%llu", ts[0].current_dist[j], ts[0].num_members[j]);
      printf("\n");
      for(j=0;j<k;j++){
	printf(" Center %d:", j);
	dumpVector(ts[0].centerVectors[j], vlen, VT_F32, 3);
      }
    }

    //Do correction on first pass for centers w/ zero  
  }



}
