/**
 * Name:  knn.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <pthread.h>


#include "common.h"
#include "lists.h"


void dumpHelp(void){
  printf("knn <options> input_filename\n"
         "    -k number of nearest neighbors to find\n"
	 "    -i/I : min/max number of input vectors to try\n"
	 "    -t/T : min/max number of threads to use\n"
	 "\n"
	 "    -C   : do NOT clear cache before each run\n"
	 "    -q   : print results in tabular form\n"
	 "    -Q   : print tabular header and exit\n"
	 "    -h   : print this help menu\n");
  exit(0);
}

void * th_search(void *arg);

//=============================================================================
void * th_search(void *arg){


  FILE *f;
  bestList_t *b;
  float *tv;
  u64 i;
  size_t rc;
  u64 spot;
  u64 cut_spot;
  int vlen;

  b = (bestList_t *)arg;

  //Shorthand
  f        = b->f;
  spot     = b->start_spot;
  cut_spot = b->start_spot + b->num_spots;
  vlen     = b->vlen;
  tv       = (float *)malloc(vlen*sizeof(float)); //Holds vector from file

  //Skip the first vector (file info) and then move on to starting vector
  rc = fseeko(f, (1+spot) * (vlen * sizeof(float)), SEEK_SET);
  if(rc)
    emsg("Tried reading %d values and got back %d on initial read\n", vlen, rc);
  
  //Scan through all training vectors in the file
  for(spot=b->start_spot; spot < cut_spot; spot++){
    rc = fread(tv, sizeof(float), vlen, f);
    if(rc!=vlen)
      emsg("Tried reading %d values and got back %d\n", vlen, rc);

    //Scan through all the vectors we're testing against
    for(b=(bestList_t *)arg; b; b = (b->is_last) ? NULL : b+1)
      compareVectors(tv,spot,b);
  }

  fclose(f);
  FREE(tv); 
  return NULL;
}
//=============================================================================
u64 launchJob(char *filename, u64 vlen, u64 tvnum, int k,
	      int num_threads, u64 svnum){

  int i,j;
  FILE *f;
  struct timeval tvstart, tvstop;
  u64 us;
  u64 spot;
  u64 spots_per_thread;
  float *sv;
  int status;
  pthread_attr_t attr;
  pthread_t *threads;
  bestList_t *bblock, *b;


  //Setup all search vectors
  //srand(time(0) + getpid());
  srand(0);


  dbg("Launch with vlen=%llu tvnum=%llu k=%d num_threads=%d svnum=%llu",
	 vlen, tvnum, k, num_threads, svnum);
  sv = (float *)malloc(svnum * vlen * sizeof(float));
  if(!sv)
    emsg("Couldn't malloc");


  for(i=0; i<svnum; i++){
    srand(0);
    for(j=0; j<vlen; j++)
      sv[i*vlen + j] =  ((float)rand()) / ((float)RAND_MAX); //Note: rand() NOT thread safe
  }

  //Figure out how the slices each thread should do
  spot = 0;
  spots_per_thread = (tvnum/num_threads);
  if(spots_per_thread<1)
    emsg("Too many threads for a small dataset");


  //Setup all the threads
  threads = (pthread_t *)  malloc(num_threads * sizeof(pthread_t));
  bblock  = (bestList_t *) malloc(num_threads * svnum * sizeof(bestList_t));
  if((!threads) || (!bblock))
    emsg("Malloc failed");


  for(i=0;i<num_threads;i++){
    b = &bblock[i*svnum + 0];
    if(i==num_threads-1)
      spots_per_thread = tvnum-spot; //Last one picks up remainder

    for(j=0; j<svnum; j++){
      initBestList( &b[j], k, vlen);
      b[j].f = (j!=0) ? NULL : safeOpenFile(NULL, filename, "r"); 
      b[j].thid         = i;
      b[j].start_spot   = spot;
      b[j].num_spots    = spots_per_thread;
      b[j].searchVector = &sv[j*vlen];
      b[j].is_last      = (j==svnum-1);
    }

    spot+=spots_per_thread;
  }

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  timerMark(0);

  //Dispatch all of the threads
  for(i=0; i<num_threads; i++){
    pthread_create( &threads[i], &attr, th_search, (void *)&bblock[i*svnum+0]);
  }

  //Merge all the threads
  for(i=0; i<num_threads; i++){
    pthread_join( threads[i], (void **)&status);

    for(j=0; j<svnum; j++){
      //printf("Obtained %d,%d:", i,j);
      //dumpBest(&bblock[i*svnum + j]);
      if(i>0){
	mergeBestList(&bblock[0*svnum + j], &bblock[i*svnum + j]); 
      }
    }
  }

  timerMark(1);
  us = timerMarkDiffUS(1,0);

  if(0){
    printf("Time %llu\n",us);
    printf("Done merging. Best results are:\n");
    for(i=0;i<svnum;i++){
      printf("[%.2d] ",i);
      dumpBest(&bblock[i]);
    }
  }


  //All results have been merged down to a single
  //list. We still need to go through the list and free each item
  for(i=0; i<svnum; i++){
    //printf("Closing %d\n",i);
    b=&bblock[i];
    FREE(b->ids);
    FREE(b->distances);
    for(j=0;j<b->k; j++)
      FREE(b->matches[j]);
  }

  //Now free the overall arrays holding everything
  FREE(sv);
  FREE(threads);
  FREE(bblock);
  return us;
}


//=============================================================================
int main(int argc, char **argv){

  int i,ich;
  int k=5;
  u64 vlen=0;   //How long vectors are
  u64 tvnum=0;  //How many training vectors we have
  u64 min_ivects=1;  //How many search vectors we have
  u64 max_ivects=0;

  char *filename=NULL;
  FILE *f;
  u64 config[2];
  int min_threads=1;
  int max_threads=0;
  u64 us;
  int table_form=0;
  int exit_now=0;
  int clear_cache=1;


  //Pick off the switches/one args
  while( (ich=getopt(argc, argv, "k:t:T:i:I:CqQXh"))!=EOF){
    switch(ich){
    case 'k': k           = atoiUnits(optarg); break;
    case 't': min_threads = atoiUnits(optarg); break;
    case 'T': max_threads = atoiUnits(optarg); break;

    case 'i': min_ivects  = atoiUnits(optarg); break; 
    case 'I': max_ivects  = atoiUnits(optarg); break; 
 
    case 'C': clear_cache = 0; break;
    case 'q': table_form  = 1; break;
    case 'Q': 
      printf("File\tK\tThreads\tInputVecs\tVecComponents\tTrainingVecs\tTime(us)\tRate(inputs*training*components/us)\n");
      break;

    case 'X': exit_now=1; break;
    case 'h':
    default:
      dumpHelp();
      exit(0);
    }
  }

  if(exit_now) exit(0);

  //Clamp down on range
  if(max_threads<=0) max_threads=min_threads;
  if(max_ivects==0)  max_ivects=min_ivects;



  //Anything else tagged on at end
  if(optind < argc){
    filename=strdup(argv[optind]);
  } 

  if((!min_ivects)||(!filename)||(k<1))
    dumpHelp();


  //Get the len/num params from file
  f = safeOpenFile(NULL, filename, "r");
  fread(config, sizeof(u64), 2, f); //Read configuration info
  vlen  = config[0];
  tvnum = config[1];
  fclose(f);


  //Everything is done in a sweep now
  int t,s;
  for(t=min_threads; t<=max_threads; t*=2){
    for(s=min_ivects; s<=max_ivects; s*=2){
      
      if(clear_cache) clearSystemDiskCache();
      
      us = launchJob(filename, vlen, tvnum, k, t, s);
      if(table_form){
	printf("%s\t%d\t%d\t%d\t%llu\t%llu\t%llu\t%.3lf\n",
	       filename,k,t,s,vlen,tvnum,us,((double)(s*tvnum*vlen))/(double)us);
	
      } else {
	printf("File: %s K: %d Threads: %d Inputs: %d VecLen: %llu TvNum: %llu Time: %llu  Rate: %.3lf\n",
	       filename,k, t,s,vlen, tvnum, us, ((double)(s*tvnum*vlen))/(double)us);
      }
    }
  }   

  return 0;
}
