/**
 * Name:  lists.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <float.h>

#include "common.h"
#include "lists.h"

//=============================================================================
void initBestList(bestList_t *b, int k, int vlen){
  int i;

  b->f            = NULL;
  b->thid         = -1;
  b->start_spot   = 0;
  b->num_spots    = 0;
  b->k            = k;
  b->vlen         = vlen;
  b->searchVector = NULL; //User passes in
  b->ids          = (u64 *)   malloc(k * sizeof(u64));
  b->distances    = (float *) malloc(k * sizeof(float));
  b->matches      = (float **)malloc(k * sizeof(float *));
  b->is_last      = 0;

  if((!b->ids) || (!b->distances) || (!b->matches))
    emsg("Malloc problem");

  for(i=0;i<k;i++){
    b->distances[i] = FLT_MAX;
    b->matches[i] = (float *)malloc(vlen*sizeof(float));
    if(!b->matches[i])
      emsg("Malloc problem");
  }

}

//=============================================================================
bestList_t * createBestList(int k, int len){
  bestList_t *b;
  b = (bestList_t *)malloc(sizeof(bestList_t));
  initBestList(b, k, len);
  return b;
}
//=============================================================================
void deleteBestList(bestList_t *a){
  int i;
  
  for(i=0;i<a->k;i++)
    FREE(a->matches[i]);
  FREE(a->matches);
  FREE(a->distances);
  FREE(a->ids);
  FREE(a);
}
//=============================================================================
void resetBestList(bestList_t *b){
  int i;
  memset(b->ids,       0, b->k*sizeof(u64));
  for(i=0;i<b->k;i++){
     b->distances[i] = FLT_MAX;
  }
}
//=============================================================================
// Fold b into a. This FREEs up malloc'd data for b, but does not FREE b 
// (or its search vector)
void mergeBestList(bestList_t *a, bestList_t *b){
  int ai,bi;
  int i;

  //printf("=====Merge Original=======\n");
  //printf("a==>\n");
  //dumpBest(a);
  //printf("b==>\n");
  //dumpBest(b);


  //printf("XXXMerging..\n");
  for(bi=ai=a->k-1; ai >= 0; ai--){
    //printf("Comparing a%d %.3f to b%d %.3f\n", ai, a->distances[ai], bi, b->distances[bi]); 
    if( b->distances[bi] < a->distances[ai] ){
      //printf("Releasing a%d\n",ai);
      //b is closer than a, insert into list
      FREE(a->matches[0]);
      for(i=0; i<ai; i++){
	//printf("filling %d\n", i);
	a->distances[i] = a->distances[i+1];
	a->ids[i]       = a->ids[i+1];
	a->matches[i]   = a->matches[i+1];
      }
      a->distances[ai] = b->distances[bi];
      a->ids[ai]       = b->ids[bi];
      a->matches[ai]   = b->matches[bi];
      bi--;
    }
  }
  while(bi>=0){
    FREE(b->matches[bi]);
    bi--;
  }

  FREE(b->ids);
  FREE(b->distances);

  //printf("=====Merge Result=======\n");
  //printf("a==>\n");
  //dumpBest(a);
  //printf("=========================================\n");

}

//=============================================================================
int compareVectors(float *a, u64 id, bestList_t *best){

  int i,j,vlen;
  float v,sum;
  float thresh;
  float *dslot;
  float *b;
  

  thresh = best->distances[0];
  vlen   = best->vlen;
  b      = best->searchVector;

  for(i=0, sum=0.0; i < vlen; i++){
    v = a[i] - b[i];
    sum += (v*v);
    if(sum>=thresh) return i+1; //bail out when we do worse than worst, best result
  }

  //Winner: find a slot and store the data
  dslot = best->matches[0]; //reuse the data slot of the previous worst value
  for(i=best->k-1; best->distances[i] < sum; i--);
  for(j=1; j<=i; j++){
    //Scoot everything up a slot
    best->ids[j-1]       = best->ids[j];
    best->distances[j-1] = best->distances[j];
    best->matches[j-1]   = best->matches[j];
  }
  //Insert the item here
  best->ids[i]       = id;
  best->distances[i] = sum;
  best->matches[i]   = dslot;
  memcpy(dslot, a, vlen*sizeof(float));

  return i; 
}
//=============================================================================
void dumpBest(bestList_t *b){
  int i;
  //printf("Best: k=%d len=%d. Best distances/ids: ", b->k, b->vlen);
  printf("Best: k=%d len=%d. %s Best distances/ids: ", b->k, b->vlen, (b->is_last) ? "(L)": "   " );
  for(i=0;i<b->k;i++)
    printf(" (%.3f/%0.4d)", (b->distances[i]!=FLT_MAX)?b->distances[i]:-1.0, b->ids[i]);
  printf("\n");

  //printf("    search is : ");
  //for(i=0;i<b->vlen;i++)
  //  printf(" %.3f", b->searchVector[i]);
  //printf("\n");

}
