/**
 * Name:  sort_mt.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64
#define _MULTI_THREADED 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <pthread.h>
#include <assert.h>

#include "common.h"




typedef enum {
  MERGE_UNKNOWN=0,
  MERGE_ROUND_ROBIN=1,
  MERGE_TREE=3
} merge_et;


//Configuration
int sort_via_pointers    = 0;
u64 cmp_word_limit       = 0; //Limit the number of words that are checked in compare
int clear_disk_cache     = 1; //clear the cache before a run
merge_et merge_algorithm = MERGE_TREE;




typedef struct {
  FILE  *fi;
  float *block;
  u64    next;
  s64    block_bytes_left;
  s64    file_bytes_left;
} merge_t;


typedef struct {
  int   id;               //Thread ID
  FILE *fi;               //Input file opened to the starting point
  u64   vects;            //number of vectors to handle
  u64   bytes_per_thread; //How much data to msort at a time
  char *intermediate_file_tag; //First part of label used for intermediate files
} thread_info_t;

//Used in log2 merger routines
typedef struct {
  int     fnum[2];       //File number for left/right
  int     completed[2];  //Complete for left/right
  float  *vdata[2];      //Pointer to data in file's cache for left/right
  int     next_node;     //Absolute id of node where this is being sent
} node_t;



//Globals
u64 vect_length = 1;
u64 vect_type   = VT_F32;
int vect_bytes  = 1*sizeof(float);



void dumpHelp(void){
  printf("sort_mt <options> [files]\n"
         " Common Options:\n"
	 "    -t/T num_threads : min/max number of threads in initial stage\n"
         "    -b/B buffer_size : min/max total amount of ram available for data\n"
	 "\n"
	 " Other Options:\n"
	 "    -1               : only do initial threaded sort\n"
         "    -2               : only do secondary merge\n"
	 "    -C               : do not clear disk cache between runs\n"
	 "    -k               : keep intermediate files\n"
	 "    -p               : use pointers during sorting (good w/ long vectors)\n"
	 "    -l vector_length : Limit comparisons to first vector_length words of a vector\n"
	 "    -Q               : print tabular header info\n"
         "\n"
         " Files:\n"
         "    -i input         : original, unsorted input file\n"
	 "    -m intermediate  : path/tag for intermediate files (eg, /tmp/foo_)\n"
	 "    -o output        : output file\n");


  exit(0);
}

void * th_search(void *arg);


//=============================================================================
static int cmpVectors(const void *x, const void *y){
  float *xx, *yy;
  int i;

  xx = (float *)x;
  yy = (float *)y;


  for(i=0; i<cmp_word_limit;i++){
    if(xx[i] < yy[i]) return -1;
    if(xx[i] > yy[i]) return  1;
    //Equals, keep going
  }
  return 0; //All were equal
}
//=============================================================================
// When cells become large enough, want to sort pointers instead of values
static int cmpVectorsViaPtrs(const void *x, const void *y){
  float *xx, *yy;
  int i;

  xx = *((float **)x);
  yy = *((float **)y);


  for(i=0; i<cmp_word_limit;i++){
    if(xx[i] < yy[i]) return -1;
    if(xx[i] > yy[i]) return  1;
    //Equals, keep going
  }
  return 0; //All were equal
}


//=============================================================================
void * th_sort(void *arg){


  FILE *fo;

  char *tmp_name;
  float *buf;
  int i,j;
  s64 num;
  float **bundle_ptrs=NULL;
  thread_info_t *ti;
  u64 vects_left;
  u64 cur_bundle;

  ti = (thread_info_t *) arg;
  vects_left = ti->vects;
  cur_bundle = ti->bytes_per_thread / vect_bytes;


  buf = (float *)mallocOrDie(ti->bytes_per_thread);
  tmp_name = (char *)mallocOrDie(2*(strlen(ti->intermediate_file_tag)+4+9));
  
  if(sort_via_pointers)
    bundle_ptrs = (float **)mallocOrDie(cur_bundle*sizeof(float *));




  for(i=0; vects_left; i++, vects_left-=cur_bundle){

    if(vects_left<cur_bundle) cur_bundle=vects_left;
  

    sprintf(tmp_name, "%s_%.3d_%.8d",ti->intermediate_file_tag,ti->id,i); 

    fo = openFileWithMetaOrDie(tmp_name, "w",
			      &vect_length, &cur_bundle, &vect_type); 

    num = fread(buf,vect_bytes, cur_bundle, ti->fi);
    if(num!=cur_bundle)
      emsg("Read of %llu vectors returned %llu", cur_bundle, num);

    //Do the actual sorting
    if(!sort_via_pointers){

      //Do the sort moving all data values around (instead of pointers)
      qsort((void *)buf, cur_bundle, vect_bytes, cmpVectors);

      //Write the results out to file
      num=fwrite(buf, vect_bytes, cur_bundle, fo);
      if(num!=cur_bundle)
	emsg("Write of %llu vectors returned %llu", cur_bundle, num);


    } else {
      //Sorting using a set of pointers

      //First need to construct a list of pointers for this block
      for(j=0;j<cur_bundle;j++)
	bundle_ptrs[j] = &buf[ j*vect_length ];

      //Sort the pointers with a special sorter function
      qsort((void *)bundle_ptrs, cur_bundle, sizeof(float *), cmpVectorsViaPtrs);
      
      //Write out all the vectors individually
      for(j=0; j<cur_bundle; j++){
	num=fwrite(bundle_ptrs[j], vect_bytes, 1, fo);
	if(num!=1)
	  emsg("Write of vector %d failed", j);
      }

    }
      
    fclose(fo);
  }

  if(sort_via_pointers) { FREE(bundle_ptrs); }
  FREE(tmp_name);
  FREE(buf);

  return NULL; //(void *)ti->id;
}
//=============================================================================
int mergeFiles(char *filename_src_base, u64 num_in_files, 
	       char *filename_dst){

  merge_t *ma;
  char *tmp_name;
  float *tmp_ptr;
  int i,j,t;
  int best_spot;
  u64    tmp_num;
  float *best_ptr;
  FILE *fo;
  s64 current_bytes;
  u64  vect_tot;
  const u64 merge_block_bytes = (4*1024);

  s64 rc;

  u64 dbg_tot_size=0;
  u64 dbg_read_in=0;
  u64 dbg_wrote_out=0;


  assert(vect_bytes < merge_block_bytes); //Make sure we have room


  tmp_name = (char *)mallocOrDie(strlen(filename_src_base)+20);

  //Create all of the merge structures
  ma=(merge_t *) mallocOrDie(num_in_files*sizeof(merge_t));

  for(i=0, t=0, vect_tot=0; i<num_in_files; ){

    for(j=0; i<num_in_files; j++){
      sprintf(tmp_name, "%s_%.3d_%.8d", filename_src_base, t,j);
      ma[i].fi = openFileWithMeta(tmp_name, "r", NULL, &tmp_num, NULL);
      dbg("Looking for intermediate file '%s'-- %s", tmp_name, (!ma[i].fi)?"Not found":"found");
      if(!ma[i].fi) break; //Done with this thread

      //File is here, add it to our list of jobs and do initial setup
      ma[i].block = (float *)mallocOrDie(merge_block_bytes);
      ma[i].next  = 0;
      ma[i].block_bytes_left = 0;
      ma[i].file_bytes_left = tmp_num * vect_bytes;
      vect_tot += tmp_num;
      dbg_tot_size += ma[i].file_bytes_left;

      i++;
    }
    t++;

    if((j==0) && (i<num_in_files)){
      emsg("Merger got fewer files than expected (%d vs %llu)",i,num_in_files);
      num_in_files=i; //bailout
    }

  }
  


  dbg("Merging %llu files, with a total of %llu vectors", num_in_files, vect_tot);
      
  //Parse through all the files, loading a block at a time
  fo = openFileWithMetaOrDie(filename_dst, "w", &vect_length, &vect_tot, &vect_type);
  do {

    if((debug) && ((dbg_wrote_out & 0x0FFFFFF)==0)) dbg("Written %llu", dbg_wrote_out);
    
    best_spot = -1;

    for(i=0; i<num_in_files; i++){

      if( (!ma[i].file_bytes_left) && (!ma[i].block_bytes_left)) continue; //this file is done
      if(!ma[i].block_bytes_left){
	current_bytes = (ma[i].file_bytes_left<merge_block_bytes) ?
	  ma[i].file_bytes_left :  //Fill in what's left
	  merge_block_bytes;       //Fill the block

	assert(!(current_bytes % vect_bytes)); //Files should have exact vectors

	rc = fread(ma[i].block, 1, current_bytes, ma[i].fi);
	if(rc!=current_bytes){
	  emsg("Read %llu bytes when expected %llu", rc, current_bytes);
	}
	dbg_read_in += current_bytes / vect_bytes;
	ma[i].next = 0;
	ma[i].block_bytes_left = current_bytes;
	ma[i].file_bytes_left -= current_bytes;
      }

      //See if this column has a better value
      if(best_spot<0){ //none selected, pick this one
	best_spot=i;
	best_ptr = &ma[i].block[ ma[i].next * vect_length];
      } else { //something selected, must compare
	tmp_ptr = &ma[i].block[ ma[i].next * vect_length];
	rc = cmpVectors( best_ptr, tmp_ptr);
	if(rc>0){
	  best_spot = i;
	  best_ptr  = tmp_ptr;
	}
      }
    }

    if(best_spot>=0){
      //Write this single vector out and advance
      rc = fwrite(best_ptr, vect_bytes, 1, fo);
      if(rc!=1) emsg("write didn't work");
      ma[best_spot].next++;
      ma[best_spot].block_bytes_left -= vect_bytes;
      dbg_wrote_out++;
    }

  } while(best_spot>=0); //ie, still finding results    

  dbg("Finished merging");

  fflush(fo);
  fclose(fo);
  sync();


  //Cleanup
  for(i=0; i<num_in_files; i++){
    FREE(ma[i].block);
  }
  FREE(ma);
  FREE(tmp_name);
  return 0;
}





//=============================================================================
// result: 0=left, 1=right, -1=completed
static int mergeFilesLog2_eval(node_t *n){

  int tst;

  tst = (n->completed[1]<<4) | n->completed[0];

  switch(tst){
  case 0x00: //both valid
    tst = cmpVectors( n->vdata[0],   n->vdata[1]);
    return (tst<=0) ? 0 : 1; 
    break;

  case 0x01: return 1; break; //left is done: right wins
  case 0x10: return 0; break; //right is done: left wins
  case 0x11: //both are done
    return -1;
    break;
  default:
    emsg("Unknown case: %x", tst);
  }

  return -2;
}

//=============================================================================
void dumpTree(node_t *n, int num_nodes){
  int i;
  int next_level, this_level;
  int log2_level;

  printf("Dump: num_nodes=%d\n",num_nodes);

  for(log2_level=0; (1<<log2_level) < num_nodes; log2_level++);
  next_level = 0;
  this_level = 0;

  for(i=0; i<num_nodes-1; i++){
    if(i==next_level){
      printf("Level %d===\n", log2_level);
      log2_level--;
      this_level = next_level;
      next_level += (1<<log2_level);
    }
    printf("[%d] (%d) Fnum:(%2d,%2d), Completed: (%2d,%2d), val: (%.3f,%.3f), next: %d (winner=%d)\n",
	   i, i-this_level,
	   n[i].fnum[0],      n[i].fnum[1],
	   n[i].completed[0], n[i].completed[1],
	   *(n[i].vdata[0]),    *(n[i].vdata[1]),
	   n[i].next_node,
	   mergeFilesLog2_eval(&n[i]));
	   
  }
  printf("DumpDone\n");
  

}
//=============================================================================
void mergeFilesLog2(char *filename_src_base, u64 num_in_files,
		    u64 total_buf_size,
		    char *filename_dst){

  typedef struct {
    FILE  *fi;
    float *block;
    u64    next;
    s64    block_vectors_left; //decrement happens when vector actually used
    s64    file_vectors_left;  //decrement happens when block is read in
    int    completed;
  } freader_t;


  freader_t *f;
  node_t *val;
  FILE *fo;
  u64 vect_bytes;

  int log2numfiles;
  int winner_lr;
  char *tmp_name;
  node_t *n;
  int i,j,t;
  u64 tmp_len, tmp_num, tmp_type;
  u64 block_size, block_num_vec;
  u64 vect_len, vect_tot, cur, num;
  u64 dbg_tot_size;
  int levelstart, levellen;
  int root;
  int rc;

  dbg("==[ Launching log2 merge sort ]=======================");


  tmp_name = (char *)mallocOrDie(strlen(filename_src_base)+20);

  if(num_in_files==1){
    //Special case: only one file, just copy it
    sprintf(tmp_name, "%s_%.3d_%.8d", filename_src_base, 0,0);
    copyFileOrDie(filename_dst, tmp_name);
    FREE(tmp_name);
    return;
  }



  //Figure out how big array needs to be
  for(log2numfiles=0; (1<<log2numfiles)<num_in_files; log2numfiles++);

  
 
  f = (freader_t *)callocOrDie(num_in_files, sizeof(freader_t));


  for(i=0, t=0, vect_tot=0; i<num_in_files; ){

    for(j=0; i<num_in_files; j++){
      sprintf(tmp_name, "%s_%.3d_%.8d", filename_src_base, t,j);
      dbg("Looking for intermediate file '%s'-- %s", tmp_name, (!fileExists(tmp_name))?"Not found":"found");
      if(!fileExists(tmp_name)) break;//Done with this thread

      //Need to find how many vectors are in file
      f[i].fi    = openFileWithMetaOrDie(tmp_name, "r", &tmp_len, &tmp_num, &tmp_type);
      if(i==0){
	//Figure dims out from first file
	vect_len = tmp_len;
	vect_bytes = vect_len*getVTypeSize(tmp_type);
	block_size = total_buf_size / num_in_files;
	block_num_vec = block_size / vect_bytes;
	block_size = block_num_vec * vect_bytes;
      }
 
      //File is here, add it to our list of jobs and do initial setup
      cur = (tmp_num<block_num_vec) ? tmp_num : block_num_vec;
      f[i].block = (float *)mallocOrDie(block_size);
      memset(f[i].block,0xFF, block_size);

      f[i].next  = 0;
      f[i].block_vectors_left = cur;
      f[i].file_vectors_left = tmp_num-cur;
      f[i].completed = 0;
      num = fread(f[i].block, vect_bytes, cur, f[i].fi);
      if(num!=cur)
	emsg("Read of %llu vectors came up with %llu", cur, num);


      vect_tot += tmp_num;
      dbg_tot_size += tmp_num*vect_bytes;

      i++;
    }
    t++;
    if((j==0) && (i<num_in_files)){
      emsg("Merger got fewer files than expected (%d vs %llu)",i,num_in_files);
      num_in_files=i; //bailout
    }
  }
  FREE(tmp_name);



  fo = openFileWithMetaOrDie(filename_dst, "w", &vect_len, &vect_tot, &tmp_type);

  //Create nodes. First level is half the number of file readers, but
  //overall size is twice that number
  //eg  log2numfiles=4: 16 file readers, 8+4+2+1 nodes=15 
  n = (node_t *)callocOrDie(1<<log2numfiles, sizeof(node_t));
  for(i=0, levelstart=0; i<log2numfiles; i++, levelstart+=levellen){ //Each level
    levellen = (1<<(log2numfiles - (i+1)));
    for(j=0; j<levellen; j++){
      
      if(i==0){
	//First level: inherit from files 
	n[levelstart+j].fnum[0]      = ((j*2+0)>=num_in_files) ? -1   :    j*2+0;
	n[levelstart+j].fnum[1]      = ((j*2+1)>=num_in_files) ? -1   :    j*2+1;
	n[levelstart+j].vdata[0]     = ((j*2+0)>=num_in_files) ? NULL : &f[j*2+0].block[0];
	n[levelstart+j].vdata[1]     = ((j*2+1)>=num_in_files) ? NULL : &f[j*2+1].block[0];
	n[levelstart+j].completed[0] = ((j*2+0)>=num_in_files) ? 1    :  f[j*2+0].completed;
	n[levelstart+j].completed[1] = ((j*2+1)>=num_in_files) ? 1    :  f[j*2+1].completed;
      }
      //All levels: find the winner
      winner_lr = mergeFilesLog2_eval( &n[levelstart+j] );
      
      
      //if(winner_lr<0)
      //  emsg("Init found node with both nodes completed?");
      
      if(i!=log2numfiles-1){ //last level
	//Move results forward
	int next_node = levelstart + levellen + j/2;
	n[levelstart+j].next_node = next_node;
	
	if(winner_lr>=0){
	  n[next_node].completed[ j&0x01 ] = 0; 
	  n[next_node].fnum[ j&0x01 ]  = n[levelstart+j].fnum[ winner_lr ];
	  n[next_node].vdata[ j&0x01 ] = n[levelstart+j].vdata[ winner_lr ]; 
	} else {
	    //This can happen in init when we have extras that don't make power of 2
	  n[next_node].completed[ j&0x01 ] = 1; 
	  n[next_node].fnum[ j&0x01 ]  = -1;
	  n[next_node].vdata[ j&0x01 ] = NULL;
	}
	
      }else {
	//last value
	n[levelstart+j].next_node = -1; //last
	  root = levelstart+j;
      }
    }
  }

  //dumpTree(n, 1<<log2numfiles);

  int round=0;

  //Keep chewing on data until everything is done
  while( (!n[root].completed[0]) || (!n[root].completed[1])){
     //Get the next winner and write out to disk
    winner_lr = mergeFilesLog2_eval( &n[root] );
    assert((winner_lr>=0) && (winner_lr<2));

    //printf("====>Round %d Winner is %d\n",round++, n[root].fnum[winner_lr]);

    rc = fwrite(n[root].vdata[winner_lr], vect_bytes, 1, fo);
    if(rc!=1) emsg("Write didn't work");
   
    //advance input file
    i = n[root].fnum[winner_lr];
    assert(!f[i].completed);
    assert( f[i].block_vectors_left ); //dec happens when va;
    f[i].block_vectors_left--;
    if(f[i].block_vectors_left){ //still in block
      f[i].next++;
    } else { //block is out of data
      
      if(f[i].file_vectors_left){
	//Grab next block of data from file
	cur = (f[i].file_vectors_left<block_num_vec) ? f[i].file_vectors_left : block_num_vec;
	f[i].next=0;
	f[i].block_vectors_left = cur;
	f[i].file_vectors_left -= cur;
	num = fread(f[i].block, vect_bytes, cur, f[i].fi);
	if(num!=cur)
	  emsg("Read of %llu vectors came up with %llu", cur, num);

      } else {
	//Done with everything, pass completed down
	f[i].completed = 1;
	n[i/2].completed[i&0x01] = 1;
      }
    }

    //update first pointer
    n[i/2].vdata[i&0x01] = &f[i].block[ f[i].next * vect_len ];

    //Follow chain down to node just before root, updating all the comparisons along the way
    for(i=i/2; n[i].next_node>=0; i=n[i].next_node){
      winner_lr = mergeFilesLog2_eval( &n[i] );
      if(winner_lr>=0){
	n[ n[i].next_node ].fnum[  i&0x01 ] = n[i].fnum[ winner_lr];
	n[ n[i].next_node ].vdata[ i&0x01 ] = n[i].vdata[winner_lr];
      }else{
	//Both parents done
	n[ n[i].next_node ].completed[ i&0x01 ] = 1;
      }
    }

    //dumpTree(n, 1<<log2numfiles);
  }
  fclose(fo);

  //Cleanup
  for(i=0;i<num_in_files;i++){
    FREE(f[i].block);
    fclose(f[i].fi);
  }
  FREE(f);
  FREE(n);


}

//=============================================================================
void launch(char *fname_in, char *fname_tmp, char *fname_out,
	    u64 total_vects,
	    int num_threads,	    
	    u64 buffer_bytes_total,
	    int do_qsort, int do_msort,
	    int keep_intermediate){

  int i,j,t;
  u64 tot_files=0;
  u64 bytes_per_thread;
  thread_info_t *tdat;
  int status;
  pthread_attr_t attr;
  pthread_t *threads;
  u64 vect_per_block, tmp, tmp_vects;
  u64 vect_per_thread;
  int rc;
  u64 time_stage1, time_stage2;

  time_stage1 = time_stage2 = 0;

  if(clear_disk_cache)
    clearSystemDiskCache();

  bytes_per_thread = buffer_bytes_total/num_threads;

  //Figure out how many files there are
  vect_per_block = bytes_per_thread / vect_bytes; //How many vectors fit in a thread's block

  //Everyone but the last one has same
  tmp_vects = total_vects / num_threads;
  tmp = tmp_vects/vect_per_block;
  if(tmp_vects % vect_per_block) tmp++; //May need one extra block
  tot_files = tmp * (num_threads-1);
  //Last file gets any extra vectors when not evenly distributable
  tmp_vects += (total_vects % num_threads); 
  tmp = tmp_vects/vect_per_block;
  if(tmp_vects % vect_per_block) tmp++; //May need one extra block
  tot_files += tmp;

  dbg("Starting.. Expecting %llu files total of %llu vects", tot_files, total_vects);

  //Stage 1: Launch threads to quick sort different sections of file
  timerMark(0);
  if(do_qsort){
    dbg("Launching %d threads doing qsort", num_threads);

    tdat    = (thread_info_t *)calloc(num_threads,sizeof(thread_info_t));
    threads = (pthread_t *)    calloc(num_threads,sizeof(pthread_t));

    vect_per_thread = total_vects / num_threads;

    for(i=0; i<num_threads;i++){
      u64 offset;
      //Create each unit
      tdat[i].id=i;
      tdat[i].bytes_per_thread = bytes_per_thread;
      tdat[i].intermediate_file_tag = fname_tmp;
      tdat[i].vects = vect_per_thread;
      if(i==num_threads-1){
	//Last guy gets any leftovers
	tdat[i].vects += (total_vects % num_threads);
      }
      offset = i*vect_per_thread;
      //Open file, then seek to this thread's work
      tdat[i].fi = openFileWithMetaOrDie(fname_in, "r", NULL, NULL, NULL); //open file, skip meta info
      rc = fseeko(tdat[i].fi, vect_bytes * (i*vect_per_thread), SEEK_CUR);
      if(rc<0)
	emsg("Couldn't seek in file");

    }

    //Prepare for launch
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

    //Dispatch all threads
    for(i=0;i<num_threads;i++){
      rc = pthread_create( &threads[i], NULL, th_sort, (void *)&tdat[i]);
    }

    //Problem: this is crashing when use i. Overflow somewhere?
    int j;
    for(j=0;j<num_threads;j++){
      int ret;
      ret = pthread_join( threads[j], (void **)&status);
      if(ret) printf("Warning? Joing had problem on %d ret: %d\n", j, ret);
    }

    for(j=0; j<num_threads;j++)
	fclose(tdat[j].fi);
 
    FREE(tdat);
    FREE(threads);
  }

  timerMark(1);

  time_stage1 = timerMarkDiffUS(1,0);
  dbg("Finished stage 1 in %llu us", time_stage1);


  //Stage 2: Merge all the results together
  if(do_msort){
    dbg("Launching merge sort on %llu files", tot_files);
    switch(merge_algorithm){
    case MERGE_ROUND_ROBIN: mergeFiles(        fname_tmp, tot_files, fname_out); break;
    case MERGE_TREE:        mergeFilesLog2(    fname_tmp, tot_files, buffer_bytes_total, fname_out); break;
    default:
      emsg("Unknown merge type?");
    }
  }
  timerMark(2);
  
  time_stage2 = timerMarkDiffUS(2,1);
  dbg("Finished stage 2 in %llu us", time_stage2);



  printf("%d\t%llu\t%llu\t%llu\t%llu\t%llu\t%llu\t%llu\t%.5lf\t%s\n",
	 num_threads,
	 buffer_bytes_total/(1024*1024),
	 total_vects, 
	 vect_length,
	 tot_files,
	 time_stage1/1000000, 
	 time_stage2/1000000, 
	 (time_stage1+time_stage2)/1000000,
	 ((double)total_vects)/((double)(time_stage1+time_stage2)),
	 fname_in);

  //Cleanup: get rid of temp files unless told to keep them
  if(!keep_intermediate){
    char *tmp_name;
    tmp_name = (char *)mallocOrDie(strlen(fname_tmp)+20);

    for(t=i=j=0; i<tot_files; ){
      sprintf(tmp_name, "%s_%.3d_%.8d", fname_tmp, t,j);
      rc=remove(tmp_name);
      if(rc<0){ //File isn't there
	if(!j) break; //didn't see any for a thread, out of luck
	//Roll over to next thread
	t++;
	j=0;
	
      } else {
	//Found a file, keep looking on this thread
	j++;
	i++;
      }
    }
    FREE(tmp_name);
  }

}

//=============================================================================
int main(int argc, char **argv){

  int i,ich;
  int t;
  u64 b;
  u64 vect_num=0;  //How many training vectors we have
  int last_time=0; //used to figure out when to delete files
  

  FILE *f;
  u64 config[2];
  int num_threads=1;
  int max_threads=1;

  u64 buffer_bytes=128*1024*1024; //How much memory we have total
  u64 buffer_max  =0;
  
  int do_qsort = 1; //Do first half: threaded qsort
  int do_msort = 1; //Do second half: sequential merge sort
  int keep_intermediate = 0; //Keep intermediate fils  


  char *fname_in  = NULL;
  char *fname_tmp = "/tmp/im";
  char *fname_out = "/tmp/sorted.dat";
  int   rc;
  int   exit_now = 0;

  //Pick off the switches/one args
  while( (ich=getopt(argc, argv, "t:T:b:B:l:i:m:o:12CpkdQXh"))!=EOF){
    switch(ich){

    case 't': num_threads  = atoiUnits(optarg); break;
    case 'T': max_threads  = atoiUnits(optarg); break;
    case 'b': buffer_bytes = atoiUnits(optarg); break; //select how much room available
    case 'B': buffer_max   = atoiUnits(optarg); break;
    case 'l': cmp_word_limit = atoiUnits(optarg); break; //limit how many words in vector are checked

    case 'i': fname_in  = strdup(optarg); break; //Get input
    case 'm': fname_tmp = strdup(optarg); break; //Get intermediate tag
    case 'o': fname_out = strdup(optarg); break; //Get output file

    case 'Q':
      printf("Threads\tTotalBufSizeMB\tNumVects\tVectLen\tIntermedFiles\tStage1(s)\tStage2(s)\tTotal(s)\tMVects/s\tFile\n");
      break;

    case 'X': exit_now = 1; break;
    case '1': do_qsort = 1; do_msort=0; break; //Only do first half
    case '2': do_qsort = 0; do_msort=1; break; //Only do second half
    case 'C': clear_disk_cache  = 0; break;
    case 'p': sort_via_pointers = 1; break;
    case 'k': keep_intermediate = 1; break; //keep intermediate results
    case 'd': debug             = 1; break;
    case 'h':
    default:
      dumpHelp();
    }
  }

  if(exit_now) exit(0);

  //Check for minimum requirements
  if(!fname_in)
    dumpHelp();

  if(max_threads<num_threads) max_threads=num_threads;
  if(num_threads<1)
    emsg("Need at least 1 thread (got %d)",num_threads);


  //Get the len/num params from file
  f = openFileWithMetaOrDie(fname_in, "r", &vect_length, &vect_num, NULL);
  fclose(f);
  vect_bytes = vect_length * sizeof(float);


  if(buffer_max<buffer_bytes) buffer_max=buffer_bytes;
  if(buffer_bytes<vect_bytes*2)
    emsg("Buffer bytes must be large enough to hold multiple vectors");

  //Word length gets set to full value if none specified
  if(!cmp_word_limit)
    cmp_word_limit=vect_length;


  //Sweep across the ranges
  last_time=0;
  for(t=num_threads; t<=max_threads; t*=2){
    for(b=buffer_bytes; b<=buffer_max; b+=buffer_bytes){

      if((t==max_threads) && (b==buffer_max))
	last_time=1;

      launch(fname_in, fname_tmp, fname_out, 
	     vect_num, t, b, 
	     do_qsort, do_msort, 
	     last_time && keep_intermediate);	
      
      //Delte sorted file unless it's the last one
      if(!last_time){
	rc = remove(fname_out);
	if(rc)
	  emsg("Problem removing file '%s'",fname_out);
      }
    }
  }


 

}
