/**
 * Name:  genVectorData.c
 *
 * Copyright (c) 2009, Lawrence Livermore National Security, LLC.
 * Produced at the Lawrence Livermore National Laboratory.
 * LLNL-CODE-416230
 * All rights reserved.
 *
 * This file is part of the Data-Intensive Benchmark Suite.
 * Please read the COPYRIGHT, LICENSE, and NOTICE files that came with this
 * distribution.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License (as published by
 * the Free Software Foundation) version 2.1 dated February 1999.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
 * conditions of the GNU Lesser General Public License for more details.
 *
 * You should have recieved a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "common.h"



void dumpHelp(void){

  printf("\n"
	 "genVectorData  -l vector_length -n number_vectors [-t u32|u64|float|double] filename\n"
	 "  -l vector_length  : how many 32b floating-point values in a vector\n"
	 "  -n number_vectors : how many vectors to put file\n"
	 "  -t data_type      : data type for vector values. Default is float\n"
	 "\n");

  exit(0);

}

int main(int argc, char **argv){

  u64 i,ich,j;
  u64 vector_length=0;
  u64 vector_num=0;
  char *filename=NULL;
  FILE *f;
  void *x;
  u64   *xi;
  u64    tot_bytes;
  u64    vector_type = VT_F32;

  struct timeval tvstart, tvstop;
  u64 us;

  //Pick off the switches/one args
  while( (ich=getopt(argc, argv, "l:n:t:h"))!=EOF){
    switch(ich){
    case 'l': vector_length = atoiUnits(optarg); break;
    case 'n': vector_num    = atoiUnits(optarg); break;
    case 't': vector_type   = getVTypeFromStr(optarg); break;
      
    case 'h':
    default:
      dumpHelp();
      exit(0);
    }
  }
    //Anything else tagged on at end
  if(optind < argc){
    filename=strdup(argv[optind]);
  } 

  if((!vector_length) || (!vector_num) || (!filename))
    dumpHelp();

  if(vector_length<1)
    emsg("Vector length has to be 1 or more");

  if(!vector_type)
    emsg("Not given a proper vector type");

  if(fileExists(filename))
    emsg("File '%s' already exists", filename);


  printf("Creating %llu vectors of length %llu and type '%s' at %s\n", 
	 vector_num, vector_length,
	 getVTypeStr(vector_type),
	 filename);

  x  = (void *)calloc(vector_length, getVTypeSize(vector_type));
  if(!x)
    emsg("Calloc problem");

  f = openFileWithMetaOrDie(filename, "w", &vector_length, &vector_num, &vector_type);

  timerMark(0);

  //Fill with data 0.0<->1.0
  for(i=0; i<vector_num; i++){
    generateRandVector(x, vector_length, vector_type);
    fwrite(x, getVTypeSize(vector_type), vector_length, f);
  }
  fclose(f);
  sync();
  timerMark(1);

  us = timerMarkDiffUS(1,0);
  tot_bytes = vector_num*vector_length;
  printf("Done. Wrote %llu bytes in %llu us (%.3lf MB/s) \n", tot_bytes, us,  (double)tot_bytes/(double)us);
  
  return 0;
}
